#include "clang.h"
#include "util.h"

extern vector <production> prods; 

clang::clang (const clang &c) {
   s = c.s;
   type = c.type;
   conf = c.conf;
   nodeId = c.nodeId; 
   neg = c.neg; 
   selfConf = c.selfConf;
   rootProd = c.rootProd;
   ps = c.ps;
   pe = c.pe;
   for (int i=0; i<c.child.size(); ++i) {
      if (c.child[i] != NULL) child.push_back(new clang(*(c.child[i])));
   }
}

void clang::operator=(const clang &c) {
   s = c.s;
   type = c.type;
   conf = c.conf;
   nodeId = c.nodeId; 
   neg = c.neg; 
   selfConf = c.selfConf;
   rootProd = c.rootProd;
   ps = c.ps;
   pe = c.pe;
      for (int i=0; i<child.size(); ++i) {
         if (child[i] != NULL) delete child[i];
      }
      child.clear();
      for (int i=0; i<c.child.size(); ++i) {
         if (c.child[i] != NULL) child.push_back(new clang(*(c.child[i])));
      }
}

void clang::giveNodeIds() {
   int id=0;
   nodeId = id;
   ++id;
   neg = false;
   for (int i=0; i<child.size(); ++i) {
      child[i]->giveNodeIds(id);
   }
}
void clang::giveNodeIds(int &id) {
   nodeId = id;
   ++id;
   neg = false;
   for (int i=0; i<child.size(); ++i) {
      child[i]->giveNodeIds(id);
   }
}
   
void clang::markNegs(const vector <int> &negIds) {
   for (int i=0; i<negIds.size(); ++i) {
      if (nodeId==negIds[i]) neg = true;
   }
   for (int i=0; i<child.size(); ++i) {
      child[i]->markNegs(negIds);
      //if (child[i]->neg) neg = true; // if child is neg then make parent neg
   }
}


bool clang::subTreeExists(const clang &subtree) const {
   // subtree exists somewhere in the tree
   //if (subTreeMatches(subtree)) return true;
   if (exactMatch(subtree)) return true;

   for (int i=0; i<child.size(); ++i) {
      if (child[i]->subTreeExists(subtree)) return true;
   }
   return false;
}
#ifdef _CLANG
bool clang::exactMatch(const clang &other) const {
   if (s != other.s/* || type!=subtree.type*/) return false;
   if (other.child.size() != child.size()) return false;

   
   bool from_goal_line=false;
   if (s=="*REGION") {
      if (child.size() > 0) {
	 if (child[0]->s=="from-goal-line") from_goal_line = true;
      }
   }

   if ((s=="*CONDITION_LIST") || (s=="*ACTION_LIST") || (s=="*REGION_LIST") || (s=="*DIRECTIVE_LIST") || (s=="*UNUM_LIST") || (s=="*ID2_LIST") || (s=="*ACTIVATION_LIST") || (s=="rec") || (s=="*CONDITION") || (from_goal_line)) {
      // order of children is not important
      vector <bool> covered(other.child.size(), false);

      for (int i=0; i<child.size(); ++i) {
	 bool found = false;
         for (int j=0; j<other.child.size(); ++j) {
	    if (! covered[j]) {
	       if (child[i]->exactMatch(*other.child[j])) {
		  found = true;
		  covered[j] = true;
		  break;
	       }
	    }
	 }
	 if (! found) {
	    return false;
	 }
      }
   }
   else {
      // order of children matters
      int start;
      if (s=="*DEFINE_RULE") start = 1; // ignore cmparing the names of the rules
      else start = 0;
      for (int i=start; i<child.size(); ++i) {
	 if (! child[i]->exactMatch(*other.child[i])) {
	    return false;
	 }
      }
   }
   return true;
}

ostream& clang::writeMRapt(ostream &out, int indent) const {
   out<<"<a>\n <mr> ";
   write(out);
   out<<" </mr>\n";
   out<<" <p> "<<rootProd<<" </p>\n</a>\n\n";

   for (int i=0; i<child.size(); ++i) {
      if (child[i]->child.size() > 0) child[i]->writeMRapt(out,indent+2);
   }

   return out;
}

ostream& clang::write(ostream &out, int indent) const {
      //for (int g=0; g<indent; ++g) out<<" ";
      if (child.size() > 0) {
	 if ((s!="*ACTION" && s!="*REGION" && child.size()==1)) {}
	 else {
	       if (child.size()==1) {
	          if (child[0]->type==KW) out<<"(";
	       }
	       else { 
		  if (s != "*PTEAM" && s!= "*PETEAM" && s!= "*PRTEAM" && s!= "*PNTEAM")
		  out<<"("; 
	       }
	 }
         //out<<"(\n";
         for (int i=0; i<child.size(); ++i) {
            if (child[i] != NULL) child[i]->write(out, indent+2);
	    else {
               //for (int g=0; g<indent+2; ++g) out<<" ";
	       out<<"NULL\n";
	    }
	    //if (i==0 && child[0]->s=="*SIDE") out<<"("; // special
	    //if (i==child.size()-1 && child[0]->s=="*SIDE") out<<")";// special
         }
         //for (int g=0; g<indent; ++g) out<<" ";
	    
	 if ((s!="*ACTION" && s!="*REGION" && child.size()==1)) {}
	 else {
	       if (child.size()==1) {
	          if (child[0]->type==KW) out<<")";
	       }
	       else { 
		  if (s != "*PTEAM" && s!= "*PETEAM" && s!= "*PRTEAM" && s!= "*PNTEAM")
		  out<<")"; 
	       }
	 }
      }
      else {
	 if ((s=="true")||(s=="false")) out<<"("<<s<<")";
	 else {
	    if (s=="direc") out<<" name "<<s<<" ";
	    else out<<s<<" ";
	 }
      }
         
      return out;
}
#endif

#ifdef _GEO
bool clang::exactMatch(const clang &other) const {
   if (s != other.s/* || type!=subtree.type*/) return false;
   if (other.child.size() != child.size()) return false;

      int start = 0;
      for (int i=start; i<child.size(); ++i) {
         if (! child[i]->exactMatch(*other.child[i])) {
            return false;
         }
      }
   return true;
}

ostream& clang::write(ostream &out, int indent) const {
   if (child.size()==0) {
      out<<s;
   }
   else {
       if (s=="*answer") {
         out<<"answer(";
         child[0]->write(out);
         out<<")\n";
      }
       else {
       if (s=="*stateid" || s=="*cityid" || s=="*lakeid" || s=="*riverid" || s=="*state_abbrvid" || s=="*placeid" || s=="*mountainid" || s=="*countryid") {
	  out<<s.substr(1, s.length())<<"(";
	       if (child.size()>0) {
		  string ss = child[0]->s;
		  for (int i=1; i<ss.size()-1; ++i) {
		    if (ss[i]=='_') {
		       if (ss[i+1]=='_') {
		          ss[i] = '.';
		          ss[i+1] = ' ';
		       }
		       else ss[i] = ' ';
		    }
		  }
		  ss = "'"+ss+"'";
		  out<<" "<<ss;
	       }
	       if (child.size()>1) {
		  if (child[1]->s != "_") out<<", '"<<child[1]->s<<"'";
		  else out<<", "<<child[1]->s;
	       }
	  out<<")";
	 }
       else {
	   child[0]->write(out);
            if (child.size() > 1) {
               out<<"(";
               for (int i=1; i<child.size(); ++i) {
                  if (i > 1) out<<", ";
                  child[i]->write(out);
               }
               out<<")";
            }
       }
       }
   }
   return out;
}
#endif

#ifdef _PC
bool clang::exactMatch(const clang &other) const {
   if (s != other.s) return false;
   if (other.child.size() != child.size()) return false;
      int start = 0;
      for (int i=start; i<child.size(); ++i) {
         if (! child[i]->exactMatch(*other.child[i])) {
            return false;
         }
      }
   return true;
}

ostream& clang::write(ostream &out, int indent) const {
   if (child.size()==0) {
      out<<s<<" ";
   }
   else {
      for (int i=0; i<child.size(); ++i) {
	 child[i]->write(out, indent+1);
      }
   }
   return out;
}
#endif

#ifdef _REST
bool clang::exactMatch(const clang &other) const {
   if (s != other.s/* || type!=subtree.type*/) return false;
   if (other.child.size() != child.size()) return false;

   if (s=="*SPL") {
      // order of children is not important
      vector <bool> covered(other.child.size(), false);
      for (int i=0; i<child.size(); ++i) {
	 bool found = false;
         for (int j=0; j<other.child.size(); ++j) {
	    if (! covered[j]) {
	       if (child[i]->exactMatch(*other.child[j])) {
		  found = true;
		  covered[j] = true;
		  break;
	       }
	    }
	 }
	 if (! found) {
	    return false;
	 }
      }
   }
   else {
      // order of children matters
      for (int i=0; i<child.size(); ++i) {
         if (! child[i]->exactMatch(*other.child[i])) {
	    return false;
         }
      }
   }
   return true;
}

ostream& clang::write(ostream &out, int indent) const {
   if (child.size()==0) {
       if (s=="rating") {
	  out<<"(rating)";
       }
       else out<<s;
   }
   else {
       if (s=="*ANSWER") {
	  for (int i=0; i<child.size(); ++i) {
	     if (i > 0) out<<"(";
	     child[i]->write(out);
	     if (i > 0) out<<")";
	  }
       }
       if (s=="*REST") {
         out<<"rest([";
         child[1]->write(out);
         out<<"])\n";
       }
       if (s=="*SPL") {
 	  for (int i=0; i<child.size(); ++i) {
	     if (i > 0 && child[i]->s=="*SPL") out<<", ";
	     child[i]->write(out);
	  }
       }
       if (s=="*RATE" || s=="*FOOD" || s=="*CITY" || s=="*STREET" || s=="*REGION" ||s=="*RESTNAME" ||s=="*COUNTY") {
	  out<<"(";
	  if (child.size()>0) {
	     string ss = child[0]->s;
	     bool quote=false;
	     for (int i=1; i<ss.size()-1; ++i) {
		if (ss[i]=='_') {
		   quote = true;
		   ss[i] = ' ';
		}
		if (ss[i]=='\'' && ss[i+1]=='\'') {
		   quote = true;
		}
	     }
	     if (quote) ss = "'"+ss+"'";
	     out<<ss;
	  }
	  out<<")";
       }
   }
   return out;
}
#endif


#ifdef _JOBS
bool clang::exactMatch(const clang &other) const {
   if (s != other.s/* || type!=subtree.type*/) return false;
   if (other.child.size() != child.size()) return false;
   
   if (s=="*SPL") {
      // order of children is not important
      vector <bool> covered(other.child.size(), false);
      for (int i=0; i<child.size(); ++i) {
	 bool found = false;
         for (int j=0; j<other.child.size(); ++j) {
	    if (! covered[j]) {
	       if (child[i]->exactMatch(*other.child[j])) {
		  found = true;
		  covered[j] = true;
		  break;
	       }
	    }
	 }
	 if (! found) {
	    return false;
	 }
      }
   }
   else {
      // order of children matters
      for (int i=0; i<child.size(); ++i) {
         if (! child[i]->exactMatch(*other.child[i])) {
	    return false;
         }
      }
   }
   return true;
}

ostream& clang::write(ostream &out, int indent) const {
   if (child.size()==0) {
       out<<s;
   }
   else {
       if (s=="*ANSWER") {
	  for (int i=0; i<child.size(); ++i) {
	     if (i > 0) out<<"(";
	     child[i]->write(out);
	     if (i > 0) out<<")";
	  }
       }
       if (s=="*JOB") {
         out<<"job([";
         child[1]->write(out);
         out<<"])\n";
       }
       if (s=="*SPL") {
 	  for (int i=0; i<child.size(); ++i) {
	     if (i==0 && child[i]->s=="*NOT") {
		out<<"not(";
		child[1]->write(out);
		out<<")";
		break;
	     }
	     else {
	        if (i==0 && child[i]->s=="salary") {
		   out<<"salary(";
		   child[1]->write(out);
		   out<<", ";
		   child[2]->write(out);
		   out<<")";
		   break;
	        }
	        else {
	           if (i==0 && child.size() > 1 && (child[i]->s=="req_exp" || child[i]->s=="des_exp")) {
		      out<<child[i]->s<<"(";
		      child[1]->write(out);
		      out<<")";
		      break;
		   }
		   else {
	              if (i > 0 && child[i]->s=="*SPL") out<<", ";
	              child[i]->write(out);
	           }
		}
	     }
	  }
       }
       if (s=="*TITLE" || s=="*LANGUAGE" || s=="*LOC" || s=="*DEGREE" || s=="*PLATFORM" ||s=="*COMPANY" ||s=="*APPLICATION" || s=="*AREA" || s=="*RECRUITER" || s=="*NUM") {
 	  for (int i=0; i<child.size(); ++i) {
	     child[i]->write(out);
	  }
       }
       if (s=="*CTITLE" || s=="*CLANGUAGE" || s=="*CLOC" || s=="*CDEGREE" || s=="*CPLATFORM" ||s=="*CCOMPANY" ||s=="*CAPPLICATION" || s=="*CAREA" || s=="*CRECRUITER") {
	  out<<"(";
	  if (child.size()>0) {
	     string ss = child[0]->s;
	     for (int i=1; i<ss.size()-1; ++i) {
		if (ss[i]=='_') {
		   ss[i] = ' ';
		}
		if (ss[i]=='\'' && ss[i+1]=='\'') {
		}
	     }
	     ss = "'"+ss+"'";
	     out<<ss;
	  }
	  out<<")";
       }
       if (s=="*CNUM" || s=="*TIME") {
	  out<<child[0]->s;
       }
   }
   return out;
}
#endif

int clang::giveRootProd() const {
   vector <string> rhs;
   for (int i=0; i<child.size(); ++i) rhs.push_back(child[i]->s);
   for (int i=0; i<prods.size(); ++i) {
      if (prods[i].getLhs()==s && prods[i].getRhs()==rhs) {
	 return i;
      }
   }
   return -1;
}

void clang::getAllRootProd() {
   rootProd = giveRootProd();
   for (int i=0; i<child.size(); ++i) {
      if (child[i] != NULL) child[i]->getAllRootProd();
   }
}

void clang::allProbs(vector <bool> &p1, vector <int> &c, vector <vector <vector <int> > > &ch) const {
   if (rootProd != -1) {
      if (p1[rootProd]) {
	 c.push_back(rootProd);
	 vector <vector <int> > hh;
         for (int i=0; i<child.size(); ++i) {
            if (child[i] != NULL && production::ifNt(child[i]->s)) {
	       vector <int> h;
	       h.push_back(child[i]->rootProd);
	       h.push_back(child[i]->ps);
	       h.push_back(child[i]->pe);
	       hh.push_back(h);
	    }
	 }
	 ch.push_back(hh);
      }
      for (int i=0; i<child.size(); ++i) {
         if (child[i] != NULL) child[i]->allProbs(p1,c,ch);
      }
   }
}

ostream& clang::writeParse(ostream &out, int indent) const {
      for (int g=0; g<indent; ++g) out<<" ";
      if (child.size() > 0) {
         out<<" ("<<s<<"  ["<<ps<<" "<<pe<<"]\n";
         //out<<"(\n";

         for (int i=0; i<child.size(); ++i) {
            if (child[i] != NULL) child[i]->writeParse(out, indent+2);
	    else {
               for (int g=0; g<indent+2; ++g) out<<" ";
	       out<<"NULL\n";
	    }
         }
         for (int g=0; g<indent; ++g) out<<" ";
         out<<")\n";
      }
      else {
         out<<s<<"\n";
      }
      return out;
}
ostream& clang::writePartition(ostream &out, const vector <string> &nl, int indent) const {
   for (int g=0; g<indent; ++g) out<<" ";
      if (child.size() > 0) {
	 //out<<rootProd<<" ";
	 //if (indent==0) {
	    //out<<"best[ ";
	    //for (int i=bestS; i<=bestE; ++i) out<<nl[i]<<" "; 
	 //}
	 out<<ps<<" "<<pe<<" "<<selfConf<<" ";
         out<<"("<<s<<"\n";
         //out<<"(\n";

	 int pi=0;
         for (int i=0; i<child.size(); ++i) {
            if (child[i] != NULL) {
	       if (production::ifNt(child[i]->s))
   {
		  for (int g=0; g<indent+2; ++g) out<<" ";
		  //out<<child[i]->ps<<" "<<child[i]->pe<<" ";
	          out<<"[ ";
	          for (int si=child[i]->ps; si<=child[i]->pe; ++si) {
	             out<<nl[si]<<" ";
	          }
	          out<<" ] "<<"\n";
		  
	          //out<<"best[ ";
	          //for (int si=child[i]->bestS; si<=child[i]->bestE; ++si) {
	            // out<<nl[si]<<" ";
	          //}
	          //out<<"]\n";
		  ++pi;
	       }
	       child[i]->writePartition(out, nl, indent+2);
	    }
	    else {
               for (int g=0; g<indent+2; ++g) out<<" ";
	       out<<"NULL\n";
	    }
         }
         for (int g=0; g<indent; ++g) out<<" ";
         out<<")\n";
      }
      else {
         out<<s<<"\n";
      }

      return out;
}
