#ifndef _CLANG_H
#define _CLANG_H

#include <string.h>
#include <iostream>
#include <fstream>
#include <vector>

#include "production.h"

using namespace std;

class clang { // MRL tree
   public:
   // data
   std::string s;
   int type;
   int nodeId;
   bool neg;
   vector <clang *> child;
   double conf, selfConf;
   int rootProd; // index of the production at the root of this tree
   int ps,pe; // start and end of coverage
   
   // functions
   clang () {} // default constructor

   clang (const std::string &as) {
      s = as;
   }
   clang (const std::string &as, int atype) {
      s = as;
      type = atype;
   }
   clang (const clang &c); // copy constructor 


   ~clang () {
      for (int i=0; i<child.size(); ++i) {
	 if (child[i] != NULL) {
	    delete child[i];
	 }
      }
   }

   void giveNodeIds();
   void giveNodeIds(int &id);
   void markNegs(const vector <int> &negIds);
   
   void operator=(const clang &c);
   bool operator == (const clang &other) const {
      return exactMatch(other);
   }
   bool operator != (const clang &other) const {
      return !exactMatch(other);
   }
   bool operator < (const clang &other) const {
      return conf > other.conf;
   }

   void makeChild(const clang &c, int i) { // make c the ith child
      if (i >= child.size()) child.resize(i+1);
      child[i] = new clang(c);
   }
   void linkChild(clang *c, int i) { // link c's clang as the ith child
      if (i >= child.size()) child.resize(i+1);
      child[i] = c;
   }
   void makeChildNull(int i) { // make ith child NULL 
      if (i >= child.size()) child.resize(i+1);
      child[i] = NULL;
   }
   void addChild(const clang &c) { // add child
      child.push_back(new clang(c));
   }
   int childSize() const {
      return child.size();
   }
   clang giveChild(int i) const{
      return *child[i];
   }

   bool subTreeExists(const clang &subtree) const;

   int giveRootProd() const;
   void getAllRootProd();

   // for a new MRL, only the following functions will need to be redefined
   bool exactMatch(const clang &other) const;
   void allProbs(vector <bool> &p1, vector <int> &c, vector <vector <vector <int> > > &ch) const;
   ostream& write(ostream &out, int indent=0) const;
   ostream& writeMRapt(ostream &out, int indent=0) const;
   ostream& writeParse(ostream &out, int indent=0) const;
   ostream& writePartition(ostream &out, const vector <string> &s, int indent=0) const;
};

inline ostream& operator << (ostream& lhs, const  clang& rhs) {
       return rhs.write(lhs);
       //return rhs.writeParse(lhs);
}

#endif
