#ifndef _EARLEY_STATE_H
#define _EARLEY_STATE_H

#include <fstream>
#include <iostream>
#include <string>
#include <vector>
#include <map>

#include "production.h"

using namespace std;

extern vector <production> prods;
extern map <string, int> nonterm;

class earleyStateIndex {
   public:
      int end, ind;

      earleyStateIndex() {
      }

      earleyStateIndex(int e, int i) {
	 end=e;
	 ind=i;
      }
};

class earleyState {

   public:
      string lhs;
      vector <string> rhs;
      int p; // production number
      double prob, selfProb;
      int begin, end, nt;
      int dot;

      vector <earleyStateIndex> coming, going;
      
      earleyState() { 
      }

      earleyState(const string &root) { // initial state
	 lhs = "";
	 rhs.push_back(root);
	 dot=0;
	 p = -1;
	 prob = 1;
	 selfProb = 1;
	 begin=0;
	 end=0;
	 nt=nonterm[lhs];
      }
  earleyState(int b, int e, int d, int pp, double conf) {
	 p = pp;
	 lhs = prods[p].getLhs();
	 rhs = prods[p].getNt(); 

	 if (prods[p].level==0) {
	    rhs = vector <string>(1,prods[p].getRhs()[0]); // just to make its size 1
	 }

	 dot = d;
	 prob = conf;
	 selfProb = conf;
	 begin=b;
	 end=e;
	 nt=nonterm[lhs];
      }
      earleyState(int b, int e, int d, int pp, const vector <string> &prhs, double conf) {
	 p = pp;
	 lhs = prods[p].getLhs();
	 rhs = prhs; 

	 dot = d;
	 prob = conf;
	 selfProb = conf;
	 begin=b;
	 end=e;
	 nt=nonterm[lhs];
      }

      bool dummyState() const {
	 return lhs=="";
      }
      
      bool complete() const {
	 return (dot==rhs.size());
      }

      bool operator==(const earleyState &other) const {
	 return (p==other.p && begin==other.begin && end==other.end && dot==other.dot && rhs==other.rhs);
      }

      ostream& write(ostream &out) const {
	 out<<lhs<<" -> ";
	 int i;
	 for (i=0; i<rhs.size(); ++i) {
	    if (dot==i) out<<" ^ ";
	    out<<rhs[i]<<" ";
	 }
	 if (dot==i) out<<" ^ ";
	 out<<"\n";
	 out<<"begin="<<begin<<" end="<<end<<" prob="<<prob<<"\n\n";
	 if (p >= 0) {
	    if (prods[p].level==0) {
	       out<<"("<<prods[p]<<")\n";
	    }
	 }
	 return out;
      }
};

inline ostream& operator << (ostream& lhs, const  earleyState& rhs) {
   return rhs.write(lhs);
}

extern int BEAM;

class earleyBeam {
   public:
   
   earleyState es;
   bool full;
   
      earleyBeam() {
	 full = false;
      }

      bool insert(const earleyState &pes) {
	 if (! full) {
	    es = pes;
	    full = true;
	    return true;
	 }
	 else {
	    if (pes.prob > es.prob) {
	       vector <earleyStateIndex> x = es.going;
	       es = pes;
	       es.going=x;
	       return true;
	    }
	    else {
	       return false;
	    }
	 }
      }
};


#endif
