#include <map>
#include <cassert>
#include <cstdlib>
#include "util.h"
#include "example.h"
#include "parse.h"

extern vector <production> prods;
extern int STR_KERNEL_LEN;
extern int MAX_MEM;
extern vector <vector <int> > nli;

/*
example::example(const vector <string> &pnl, const clang &pc, int pid, int prodSize, const ctree &s, const ctree &a, const ctree &m) { 
   // train example
   trainEg = true;
   y = vector <double> (prodSize, -1); // 0 for adatron, -1 for perceptron
   nl = pnl;
   c = pc;
   id = pid;
   getProds(c);
   
   for (int i=0; i<nl.size(); ++i) {
      nli.push_back(getS2N(nl[i]));
   }

   syn = s;
   augsyn = a;
   mrapt = m;
   pt.resize(nl.size(), vector <bool>(nl.size(), true));
}*/


// for mult
example::example(const vector <string> &pnl, const vector <clang> &mpc, int pid, int prodSize, const ctree &s, const ctree &a, const ctree &m) { 
   // train example
   trainEg = true;
   eval = false;
   y = vector <double> (prodSize, -1); // 0 for adatron, -1 for perceptron
   nl = pnl;
   mc = mpc;
   if (mc.size() > 0) c = mc[0];
   id = pid;
   for (int i=0; i<mc.size(); ++i) getProds(mc[i]);

   for (int i=0; i<nl.size(); ++i) {
      nli.push_back(getS2N(nl[i]));
   }

   syn = s;
   augsyn = a;
   mrapt = m;
   pt.resize(nl.size(), vector <bool>(nl.size(), true));
   w = 1;
}

example::example(const vector <string> &pnl, int pid, int prodSize, const ctree &s, const ctree &a, const ctree &m) { 
   // test example
   trainEg = false;
   eval = false;
   nl = pnl;
   id = pid;
   y.resize(prodSize);
   for (int i=0; i<nl.size(); ++i) {
      nli.push_back(getS2N(nl[i]));
   }
   
   syn = s;
   augsyn = a;
   mrapt = m;
   pt.resize(nl.size(), vector <bool>(nl.size(), true));
   w = 1;
}
example::example(const vector <string> &pnl, const vector <clang> &mpc, int pid, int prodSize, double pconf) { 
   // test example
   trainEg = false;
   eval = false;
   nl = pnl;
   mc = mpc;
   if (mc.size() > 0) {
      c = mc[0];
   }
   id = pid;
   y.resize(prodSize);
   for (int i=0; i<nl.size(); ++i) {
      nli.push_back(getS2N(nl[i]));
   }
   
   conf = pconf;
   pt.resize(nl.size(), vector <bool>(nl.size(), true));
   w = 1;
}

void example::resetMap() {
   k.clear();
}
 
void example::initMap() {
   if (selfK.size()==0) {
   selfK.resize(nli.size(), vector <double> (nli.size(),0));
   for (int i=0; i<nli.size(); ++i) {
      for (int j=i; j<nli.size(); ++j) {
	 vector <int> s;
	 for (int e=i; e<=j; ++e) s.push_back(nli[e]);
	 selfK[i][j] = ncsUn(s,s,STR_KERNEL_LEN);
      }
   }
   }
   k.clear();
   //k.resize(nli.size(), vector <map <int, double> > (nli.size()));
}
 
double example::giveK(int s, int e, int i, const vector <int> &n) {
   if (k.size() > MAX_MEM) k.clear();
   int ind = s*10000000 + e*100000 + i;
   if (ind < 0) {
      cout<<s<<" "<<e<<" "<<i<<" int size exceeded, example.cpp\n";
      exit(1);
   }
   if (i >= 100000) {
      cout<<i<<" nli.size() exceeded\n";
      exit(1);
   }
   hash_map <int, double>::iterator itr = k.find(ind);
   if (itr==k.end()) {
      //vector <int> st;
      //for (int a=s; a<=e; ++a) st.push_back(nli[a]);
      double d = ncsUn(n, 0, n.size()-1, nli, s, e, STR_KERNEL_LEN);
      //double d = ncsUn(n, st, STR_KERNEL_LEN);
      k[ind] = d;
      return d;
   }
   else {
      return itr->second;
   }
}
/*
double example::giveK(int s, int e, int i, const vector <int> &n) {
   if (k[s][e].find(i) == k[s][e].end()) {
      vector <int> st;
      for (int a=s; a<=e; ++a) st.push_back(nli[a]);
      k[s][e][i] = ncsUn(n, st, STR_KERNEL_LEN);
   }
   return k[s][e][i];
}*/

void example::getProds(clang &pc, int parent, int childNo) {
   string lhs = pc.s;
   vector <string> rhs;

   if (parent==-1) {
      y.resize(prods.size(), -1);
   }

   for (int i=0; i<pc.child.size(); ++i) {
      rhs.push_back(pc.child[i]->s);
   }

   int thisParent=-1;
   for (int i=0; i<prods.size(); ++i) {
      if (prods[i].getLhs()==lhs && prods[i].getRhs()==rhs) {
	 y[i] = 1;
	 pc.rootProd = i;
	 thisParent = 0;
      }
   }

   if (thisParent != -1) {
      for (int i=0; i<pc.child.size(); ++i) getProds(*pc.child[i], thisParent, i);
   }
}


void example::getProdProbs(const clang &pc, vector <double> &conf, int parent, int childNo) {
   string lhs = pc.s;
   vector <string> rhs;

   if (parent==-1) {
      conf.resize(prods.size(), 0);
   }

   for (int i=0; i<pc.child.size(); ++i) {
      rhs.push_back(pc.child[i]->s);
   }

   int thisParent=-1;
   for (int i=0; i<prods.size(); ++i) {
      if (prods[i].getLhs()==lhs && prods[i].getRhs()==rhs) {
	 if (pc.conf > conf[i])
	    conf[i] = pc.conf;
	 thisParent = 0;
      }
   }

   if (thisParent != -1) {
      for (int i=0; i<pc.child.size(); ++i) getProdProbs(*pc.child[i], conf, thisParent, i);
   }
}


#ifdef _CLANG
ostream& example::write(ostream &out) const {
   out<<"<example id=\""<<id<<"\">\n<nl>\n";
   for (int i=0; i<nl.size(); ++i) out<<nl[i]<<"  ";
   out<<"\n</nl>\n";
   if (!trainEg) {
      for (int i=0; i<ans.size(); ++i) {
         out<<"<conf> "<<ans[i].conf<<" </conf>\n";
         out<<"<iclang n=\"0\">\n";
	 if (ans[i].s != "*DEFINE_TOKEN") {
	    out<<"partial("<<ans[i]<<")\n";
	 }
	 else { 
	    out<<ans[i]<<"\n";
	 }
         out<<"</iclang>\n";
	 if (ans[i].s=="*DEFINE_TOKEN") break;
      }
   }
   else {
      out<<"<conf> "<<c.conf<<" </conf>\n";
      out<<"<iclang n=\"0\">\n";
      out<<c<<"\n";
      out<<"</iclang>\n";
   }
   out<<"</example>\n\n";
   return out;
}
ostream& example::writePartition(ostream &out) const {
   //out<<"<example id=\""<<id<<"\">\n<nl>\n";
   out<<"\n";
   for (int i=0; i<orig_nl.size(); ++i) out<<orig_nl[i]<<"  ";
   out<<"\n</nl>\n";
   if (!trainEg) {
      for (int i=0; i<ans.size(); ++i) {
         out<<"<conf> "<<ans[i].conf<<" </conf>\n";
         out<<"<iclang n=\"0\">\n";
         ans[i].writePartition(out, nl);
	 out<<"\n";
         out<<"</iclang>\n";
	 break;
      }
   }
   else {
      out<<"<conf> "<<c.conf<<" </conf>\n";
      out<<"<iclang n=\"0\">\n";
      out<<c<<"\n";
      out<<"</iclang>\n";
   }
   //out<<"</example>\n";
   return out;
}

ostream& example::writeModSyn(ostream &out) const {
   out<<"<example id=\""<<id<<"\">\n<nl>\n";
   for (int i=0; i<orig_nl.size(); ++i) out<<orig_nl[i]<<"  ";
   out<<"\n</nl>\n";
   out<<"<syn>\n"<<modSyn<<"\n</syn>\n";
   out<<"<iclang n=\"0\">\n"<<c<<"\n</iclang>\n";
   out<<"</example>\n\n";
   return out;
}
#endif

#ifndef _CLANG
ostream& example::write(ostream &out) const {
   out<<"<example id=\""<<id<<"\">\n<nl>\n";
   for (int i=0; i<nl.size(); ++i) out<<nl[i]<<"  ";
   out<<"\n</nl>\n";
   if (eval) {
      out<<"<conf> "<<conf<<" </conf>\n";
      out<<"<sem> "<<c<<" </sem>\n";
      out<<"</example>\n\n\n";
      return out;
   }
   if (!trainEg) {
      for (int i=0; i<ans.size(); ++i) {
         out<<"<conf> "<<ans[i].conf<<" </conf>\n";
         out<<"<sem>\n";
#ifdef _GEO
	 if (ans[i].s != "*answer") {
	    out<<"partial("<<ans[i]<<")\n";
	 }
	 else { 
	    out<<ans[i]<<"\n";
	 }
#endif
#ifndef _GEO
	 out<<ans[i]<<"\n";
#endif
         out<<"</sem>\n";
	 if (ans[i].s=="*answer") break;
      }
   }
   else {
      //out<<"<conf> "<<c.conf<<" </conf>\n";
      out<<"<sem>\n";
      out<<c<<"\n";
      out<<"</sem>\n";
   }
   out<<"</example>\n\n";
   return out;
}

ostream& example::writePartition(ostream &out) const {
   //out<<"<example id=\""<<id<<"\">\n<nl>\n";
   for (int i=0; i<nl.size(); ++i) out<<orig_nl[i]<<"  ";
   out<<"\n</nl>\n";
   if (!trainEg) {
      for (int i=0; i<ans.size(); ++i) {
         out<<"<conf> "<<ans[i].conf<<" </conf>\n";
         out<<"<sem>\n";
         ans[i].writePartition(out, nl);
	 out<<"\n";
         out<<"</sem>\n";
	 break; // for only the first answer
      }
   }
   else {
      out<<"<conf> "<<c.conf<<" </conf>\n";
      out<<"<sem>\n";
      //out<<c<<"\n";
      ans[0].writePartition(out, nl);
      out<<"</sem>\n";
   }
   //out<<"</example>\n";
   return out;
}
ostream& example::writeModSyn(ostream &out) const {
   out<<"<example id=\""<<id<<"\">\n<nl>\n";
   for (int i=0; i<orig_nl.size(); ++i) out<<orig_nl[i]<<"  ";
   out<<"\n</nl>\n";
   out<<"<syn>\n"<<modSyn<<"\n</syn>\n";
   out<<"<sem>\n"<<c<<"\n</sem>\n";
   out<<"</example>\n\n";
   return out;
}
#endif
/*
void example::processMrapt(ctree &mr) {
   // this will find the productions at the root of MRs (rootMRProds)  
   string m;
   bool in=false;
   for (int i=0; i<mr.root.length(); ++i) {
      if (in && mr.root[i]==']') {
	 in=false;
	 break;
      }
      if (in) m.push_back(mr.root[i]);
      if (mr.root[i]=='[') {
	 in=true;
      }
   }
   bool a=true;
   for (int i=0; i<m.length(); ++i) {
      if (m[i]=='_') m[i] = ' ';
      if (m[i]=='{') m[i] = '(';
      if (m[i]=='}') m[i] = ')';
      if (m[i]=='\'' && a) {
	 m[i] = '{';
	 a=false;
      }
      if (m[i]=='\'' && !a) {
	 m[i] = '}';
	 a=true;
      }
   }

   if (m.length()  > 0) {
      //cout<<m<<" =\n";
      ofstream tmp("/tmp/rjkate/tmp");
      tmp<<m<<"\n";
      tmp.close();
      FILE *tfile = fopen("/tmp/rjkate/tmp", "r");
      extern FILE* yyin;
      extern TOKEN parseresult;
      yyin = tfile;
      parseresult = NULL;
      yyparse();
      TOKEN p;
      p = parseresult;
      fclose(tfile);
      if (parseresult==NULL) {
	 //cout<<"NO! "<<m<<"\n";
	 mr.rootMRProd = -1;
      }
      else {
	 //cout<<"YES! "<<m<<"\n";
	 clang cl(p);
	 if (prods[cl.giveRootProd()].level==0) {
	    mr.rootMRProd = cl.giveRootProd();
	 }
	 else {
	    mr.rootMRProd = -1;
	 }
	 //if (mr.rootMRProd != -1) cout<<prods[mr.rootMRProd]<<"\n";
	 //cout<<cl<<"\n";
      }
   }
   else {
      mr.rootMRProd = -1;
   }

   for (int c=0; c<mr.child.size(); ++c) processMrapt(*mr.child[c]);
}*/

class twoInt {
   public:
      int i, ps;
      bool operator < (const twoInt &other) const {
	 return ps < other.ps;
      }
};

void example::makeSyn(ctree &t, const ctree &s, clang &l) {
   t.root = s.firstAncestor(l.ps, l.pe);
   if (prods[l.rootProd].level > 0) {
      vector <twoInt> ti;
      for (int i=0; i<l.child.size(); ++i) {
         if (production::ifNt(l.child[i]->s)) {
	    twoInt ttii;
	    ttii.i = i;
	    ttii.ps = l.child[i]->ps; 
	    ti.push_back(ttii);
         }
      }
      sort(ti.begin(), ti.end());
      for (int i=0; i<ti.size(); ++i) {
	 t.child.push_back(new ctree());
	 makeSyn(*t.child[t.child.size()-1], s, *l.child[ti[i].i]);
      }
   }
   else {
      for (int i=l.ps; i<=l.pe; ++i) {
	 ctree a = s.leafCover(i,i);
	 t.child.push_back(new ctree(a));
      }
   }
}

void example::modifySyn() {
   if (ans.size() > 0) makeSyn(modSyn, syn, ans[0]);
   else modSyn = syn;
}

