#ifndef _EXAMPLE_H
#define _EXAMPLE_H

#include <string>
#include <vector>
#include <fstream>
#include <hash_map.h>

#include "util.h"
#include "clang.h"
#include "production.h"
#include "tree.h"
#include "common_seq.h"

class example {

   vector <vector <bool> > pt; // partition true or not
 
   public:
   int id;
   clang c;
   vector <clang> mc; // multiple MRs
   
   bool trainEg;
   vector <string> nl;
   vector <int> nli;
   vector <string> orig_nl;
   vector <int> orig_nli;
   ctree syn, augsyn, modSyn, mrapt;

   //vector <vector <map <int, double> > > k; // map of kernel values
   hash_map <int, double> k;
   vector <vector <double> > selfK; // self kernels

   bool skip;
   double replConf;
   double w; // weight
   double conf;
   bool eval;

   vector <double> y; // presence of productions

   vector <bool> likelyConsts; // presence of const productions

   vector <clang> ans;

   example() {
      eval=false;
   }

   void resetMap();
   void initMap();
   double giveK(int s, int e, int i, const vector <int> &n);
   
   example(const vector <string> &pnl, const clang &pc, int pid, int prodSize, const ctree &s, const ctree &a, const ctree &m); 
   // for mult  
   example(const vector <string> &pnl, const vector <clang> &mpc, int pid, int prodSize, const ctree &s, const ctree &a, const ctree &m); 
      
   example(const vector <string> &pnl, int pid, int prodSize, const ctree &s, const ctree &a, const ctree &m);
   example(const vector <string> &pnl, const vector <clang> &mpc, int pid, int prodSize, double conf);
      
   bool ispt(int i, int j) const {
      return pt[i][j];
   }

   void getProds(clang &, int parent=-1, int childNo=-1);
   void getProdProbs(const clang &pc, vector <double> &conf, int parent=-1, int childNo=-1);

   ctree synCover(int s, int e) const {
      // returns the syntactic subtree covering words from s to e
      return syn.leafCover(s,e);
   }

   void setLabel(int i, double l) {
      y[i]=l;
   }
   void setNl(int i, const string & s) {
      assert(i < nl.size());
      nl[i] = s;
   }
   double label(int i) const {
      return y[i];
   }

   int giveId() const {
      return id;
   }

   int length() const {
      return nl.size();
   }

   const vector <string> & giveNl() const { 
      return nl;
   }
   const vector <double> & giveLabels() const { 
      return y;
   }

   string nous(const string &s) const { // no underscore
      string r=s;
      if (r[0] !='_') for (int i=0; i<r.size(); ++i) if (r[i]=='_') r[i]=' ';
      return r;
   }

   bool operator < (const example &other) const {
      if (eval) return conf > other.conf;
      if (ans.size() > 0 && other.ans.size() > 0) {
	 return ans[0].conf > other.ans[0].conf;
      }
      else {
	 return (other.ans.size()==0); 
      }
   }

   ostream& write(ostream &out) const;
   ostream& writeFormal(ostream &out) const;
   ostream& writePartition(ostream &out) const;
   ostream& writeModSyn(ostream &out) const;

   void processMrapt(ctree &);

   void modifySyn();
   void makeSyn(ctree &t, const ctree &s, clang &l);
};

inline ostream& operator << (ostream& out, const  example& t) {
       return t.write(out);
}

#endif
