#include "mrlParser.h"

void mrlParser::extractTokens() {
   for (int p=0; p<pr.size(); ++p) {
      vector <string> rhs = pr[p].getRhs();
      for (int r=0; r<rhs.size(); ++r) {
	 if (! production::ifNt(rhs[r])) {
	    if (find(token.begin(), token.end(), rhs[r]) == token.end()) {
	       token.push_back(rhs[r]);
	    }
	 }
      }
   }
   //for (int i=0; i<token.size(); ++i) cout<<"|"<<token[i]<<"|\n";
}

void mrlParser::readMrl(ifstream &in, clang &cl) const {
   vector <string> st;

   vector <char> ch;
   bool b=false;
   in>>noskipws;
   char c;
   in>>c;
   while (c==' '||c=='\t'||c=='\n') in>>c;
   while (!b) {
      if (c != '<') {
	 if (c != '\n') ch.push_back(c); // todo: improve
      }
      else {
	 in>>c>>c>>c>>c>>c;
	 b=true;
      }
      in>>c;
   }
   in>>skipws;

   //writeVector(cout,ch);

   vector <string> t;
   tokenize(ch,t);

   //writeVector(cout,t);
   
   earleyParse(t,cl);
}

void mrlParser::tokenize(const vector <char> &ch, vector <string> &t) const {
   t.clear();
   int i=0;
   while (i<ch.size()) {
      bool b=false;
      int c=0;
      while (!b && c+i < ch.size()) {
	 //cout<<c+i<<" "<<ch.size()<<"\n";
         string s;
         for (int j=i; j+c<ch.size(); ++j) s.push_back(ch[j]);
	 while (s[s.size()-1]==' '||s[s.size()-1]=='\t') s.erase(s.size()-1);
	 //cout<<s<<"\n";
         //for (int e=0; e<token.size(); ++e) cout<<"|"<<token[e]<<"| |"<<s<<"|\n";
         if (find(token.begin(), token.end(), s)!=token.end()) {
	    b=true;
	    i = ch.size()-c;
	    t.push_back(s);
	 }
	 else ++c;
      }
      if (!b && i!=ch.size()) {
	 cout<<"Error: Could not tokenize: ";
	 writeVector(cout,ch);
	 exit(1);
      }
   }
}
/*
void mrlParser::tokenize(const vector <char> &ch, vector <string> &t) const {
   string s="";
   t.clear();
   for (int i=0; i<ch.size(); ++i) {
      s.push_back(ch[i]);
      if (find(token.begin(), token.end(), s)!=token.end()) {
	 t.push_back(s);
	 s = "";
      }
   }
   assert(s=="");
}*/

bool incomplete(const estate &s) {
   return (s.dot < s.rhs.size());
}
bool nextIsTerm(const estate &s) {
   return !production::ifNt(s.rhs[s.dot]);
}
bool enqueue(const estate s, vector <estate> &ves) {
   if (find(ves.begin(), ves.end(), s)==ves.end()) {
      ves.push_back(s);
      return true;
   }
   else return false;
}

void mrlParser::predictor(estate s, vector <vector <estate> > &es) const {
   //cout<<"Inside predictor of "<<s<<"\n";
   for (int p=0; p<pr.size(); ++p) {
      if (pr[p].getLhs()==s.rhs[s.dot]) {
	 estate n(p,0,s.dotPos,s.dotPos,pr[p].getLhs(),pr[p].getRhs());
	 bool b = enqueue(n, es[s.dotPos]);
	 //if (b) cout<<"   From predictor: "<<n<<"\n";
      }
   }
}

void mrlParser::scanner(estate s, vector <vector <estate> > &es, const vector <string> &t) const {
   //cout<<"Inside scanner of "<<s<<"\n";
   if (s.rhs[s.dot]==t[s.dotPos]) {
      estate n(s.pn,s.dot+1,s.start,s.dotPos+1,s.lhs,s.rhs);
      n.previ=s.previ;
      n.prevj=s.prevj;
      bool b = enqueue(n,es[s.dotPos+1]);
      //if (b) cout<<"   From scanner: "<<n<<"\n";
   }
}

void mrlParser::completer(int i, int j, vector <vector <estate> > &es) const {
   estate s=es[i][j];
   //cout<<"Inside completer of "<<s<<"\n";
   for (int a=0; a<es[s.start].size(); ++a) {
      estate  e = es[s.start][a];
      if (incomplete(e) && e.rhs[e.dot]==s.lhs) {
	 estate n(e.pn, e.dot+1, e.start, s.dotPos, e.lhs, e.rhs);
	 n.previ = e.previ;
	 n.prevj = e.prevj;
	 n.previ.push_back(i);
	 n.prevj.push_back(j);
	 bool b = enqueue(n, es[s.dotPos]);
         //if (b) cout<<"   From completer: "<<n<<"\n";
      }
   }
}

void recoverCl(const estate &s, const vector <vector <estate> > &es, clang &cl) {
   cl.s = s.lhs;
   //cout<<"Recover: "<<s<<"\n";
   cl.rootProd = s.pn;
   int c=0;
   for (int i=0; i<s.rhs.size(); ++i) {
      if (production::ifNt(s.rhs[i])) {
	 clang ch;
	 //cout<<c<<" "<<s.previ.size()<<"\n";
	 recoverCl(es[s.previ[c]][s.prevj[c]], es, ch);
	 ++c;
	 cl.makeChild(ch,i);
      }
      else {
	 //cout<<s.rhs[i]<<"\n";
	 cl.makeChild(clang(s.rhs[i]),i);
      }
   }
}


bool mrlParser::recover(const vector <vector <estate> > &es, clang &cl) const {
   int a = es.size()-1;
   for (int i=0; i<es[a].size(); ++i) {
      if (!incomplete(es[a][i])) {
	 recoverCl(es[a][i],es,cl);
	 return true; // only one parse 
      }
   }
   return false;
}
   
void mrlParser::earleyParse(const vector <string> &t, clang &cl) const {
   vector <vector <estate> > es(t.size()+1);

   // initialize with the dummy start state
   es[0].push_back(estate(-1,0,0,0,"",vector <string>(1,startSymbol)));

   for (int w=0; w<t.size(); ++w) {
      for (int e=0; e<es[w].size(); ++e) {
	 if (incomplete(es[w][e])) {
	    if (!nextIsTerm(es[w][e])) {
	       predictor(es[w][e],es);
	    }
	    else {
	       scanner(es[w][e],es,t);
	    }
	 }
	 else {
	    completer(w,e,es);
	 }
      }
   }

   bool b = recover(es,cl);
   if (b) {
      //cl.writeParse(cout);
      //cout<<"\n";
   }
   else {
      cout<<"Error: Could not parse: ";
      writeVector(cout,t);
      exit(1);
   }
}
