#ifndef _MRLPRSER_H
#define _MRLPRSER_H

#include <fstream>
#include <iostream>
#include <vector>
#include <string>
#include <cassert>

#include "util.h"
#include "production.h"
#include "clang.h"

using namespace std;

class estate {
   public:
      int pn, dot, start, dotPos;
      string lhs;
      vector <string> rhs;

      vector <int> previ, prevj; 

      estate(int ppn, int pdot, int pstart, int pdotPos, const string &plhs, const vector <string> &prhs) {
	 pn = ppn;
	 dot = pdot;
	 start = pstart;
	 dotPos = pdotPos;
	 lhs = plhs;
	 rhs = prhs;
      }

      bool operator==(const estate &other) const {
	 return pn==other.pn && dot==other.dot && start==other.start && dotPos==other.dotPos;
      }

      ostream &write(ostream &out) const {
	 out<<lhs<<" -> ";
	 for (int i=0; i<rhs.size(); ++i) {
	    if (i==dot) cout<<" @ ";
	    out<<rhs[i];
	 }
	 out<<"[ "<<start<<", "<<dotPos<<"] ";
	 cout<<previ.size()<<"\n";
	 return out;
      }
};

inline ostream& operator<<(ostream &out, const estate &es) {
   return es.write(out);
}
class mrlParser {
   string startSymbol;
   vector <production> pr;

   vector <string> token;
   void extractTokens();

   void tokenize(const vector <char> &ch, vector <string> &t)const;

   void scanner(estate s, vector <vector <estate> > &es, const vector <string> &t) const; 
   void predictor(estate s, vector <vector <estate> > &es) const; 
   void completer(int i, int j, vector <vector <estate> > &es) const; 
   bool recover(const vector <vector <estate> > &es, clang &cl) const;
   void earleyParse(const vector <string> &t, clang &cl) const;

   public:

      mrlParser(const vector <production> &prods, const string &s) {
	 pr = prods;
	 startSymbol = s;
	 extractTokens();
      }
   
      void readMrl(ifstream &in, clang &cl) const;
      // reads an MRL till </sem> is found, assumees <sem> has been already read
};

#endif
