#include "multEg.h"
#include "hungarian.h"

//////////////////////////////////////// parameters
int MAX_ITR=5;  // maximum unmber of iterations
double CONVG=0.95; // convergence => 95% same in the next iteration
int MAX_FIRST_ITR_EGS=500; // maximum examples in the first iteration
extern bool OUT; // trace or not
extern double THR1, THR2; // confidence thresholds for including an example, 
                          // first and subsequent iterations 
////////////////////////////////////////

extern string KDIR, START;
extern int ANN; // noise parameter
bool DEBUG;
extern string outdir;
vector <clang> mc;

void writeOut(const vector <example> &testEg, const string &outFile);

void replaceConstWords(vector <example> &eg);

void readMultEgs(vector <multEg> &megs, const string &filename, const mrlParser &mp) {
   ifstream in(filename.c_str());
   assert(in);

   string s;
   while (s != "<examples>") {
      in>>s;
   }
   while (true) {
      // read examples
      string sid;
      while ((s != "</examples>") && (s != "<example")) {
	 in>>s;
      }
      if (s == "</examples>") break;
      in>>s;
      for (int i=4; i<s.length()-2; ++i) {
	 sid.push_back(s[i]);
      }
      int id = atoi(sid.c_str());
    
      in>>s;
      assert(s=="<nl>");
      vector <string> nl;
      in>>s;
      while (s != "</nl>") {
	 nl.push_back(s);
         in>>s;
      }
      //removeStopWords(nl);
      in>>s;

      assert(s=="<semid>");
      vector <int> mci;
      in>>s;
      while (s != "</semid>") {
	 mci.push_back(atoi(s.c_str()));
	 in>>s;
      }
      in>>s;
      
      if (s=="</example>") {
	 megs.push_back(multEg(id,nl,mci));
      }
   }
   while (s != "<sems>") {
      in>>s;
   }
   while (true) {
      while ((s != "</sems>") && (s != "<sem>") && (s != "<sem")) {
	 in>>s;
      }
      if (s == "</sems>") break;
      if (s == "<sem") in>>s;

      clang clg;
      mp.readMrl(in,clg);
      in>>s;
      //assert(s=="</sem>");
      /*
      ofstream tmp((KDIR+"tmp").c_str());
      s = readWord(file);
      while (s != "</sem>") {
	 tmp<<s<<" ";
	 s = readWord(file);
      }
      tmp.close();
      //string ss = "cp -f " + KDIR+"tmp " + KDIR + "tmp2";
      //system(ss.c_str());
      FILE *tfile = fopen((KDIR+"tmp").c_str(), "r");
      TOKEN clg;
      readClangTree(tfile, clg);
      fclose(tfile);
      s = readWord(file);
      */
      mc.push_back(clg);
   }

   // some padding in the end (remove later)
   for (int i=0; i<10; ++i) mc.push_back(mc[mc.size()-1]);
   in.close();
}

ostream & multEg::write(ostream &out) const {
   out<<"<example id=\""<<id<<"\">\n";
   out<<"<nl>\n";
   for (int j=0; j<nl.size(); ++j) out<<nl[j]<<" ";
   out<<"</nl>\n";
   out<<"<sem>\n";
   for (int i=0; i<mci.size(); ++i) {
      //out<<mci[i]<<" ";
      out<<mc[mci[i]]<<"\n";
   }
   out<<"</sem>\n";
   out<<"</example>\n\n";
   return out;
}

example makeExample(const vector <string> &nl, const clang &c) {
   ctree s,a,m;
   //cout<<c<<"\n";
   return example(nl, vector <clang>(1,c), 0, prods.size(),s,a,m);
}

void generateEgs(const vector <multEg> &megs, vector <example> &trainEg, vector <vector <int> > &e) {
   e.clear();
   e.resize(megs.size());
   for (int i=0; i<megs.size(); ++i) {
      for (int j=0; j<megs[i].mci.size(); ++j) {
	 example eg(makeExample(megs[i].nl, mc[megs[i].mci[j]]));
	 eg.w = 1/megs[i].mci.size();
	 //eg.w = 1;
	 trainEg.push_back(eg);
	 e[i].push_back(megs[i].mci[j]);
      }
   }
   replaceConstWords(trainEg);
}

void generateBestEgs(const vector <multEg> &megs, const semanticParser &sp, vector <example> &trainEg, vector <vector <int> > &e, double THR=0) {
   e.clear();
   e.resize(megs.size());
   //cout<<"here 1\n";
   vector <vector <example> > eg(megs.size());
   vector <egId> eid;
   for (int i=0; i<megs.size(); ++i) {
      eg[i].resize(megs[i].mci.size());
      for (int j=0; j<megs[i].mci.size(); ++j) {
	 //cout<<i<<" "<<j<<"\n";
	 eg[i][j] = makeExample(megs[i].nl, mc[megs[i].mci[j]]);
	 eid.push_back(egId(i,j));
	 vector <bool> rec(prods.size(), true);
         for (int p=0; p<prods.size(); ++p) {
	    rec[p] = true;
	    if (eg[i][j].y[p] <= 0) {
	       rec[p] = false;
	    }
         }
	 sp.earleyParses(START, eg[i][j], rec, 1, 0, true);
	 if (eg[i][j].ans.size() > 0) {
	    eid[eid.size()-1].conf = eg[i][j].ans[0].conf;
	 }
	 else {
	    eid[eid.size()-1].conf = 0;
	 }
	 //cout<<i<<" "<<j<<" --\n";
      }
   }
   //cout<<"here 2\n";

   /*   
   // uses greedy assignment
   sort(eid.begin(), eid.end());
   vector <bool> mcb(mc.size(), true);
   vector <bool> megsb(megs.size(), true);

   for (int i=0; i<eid.size(); ++i) {
      if (eid[i].conf > 0) {
         int m = megs[eid[i].nli].mci[eid[i].ci];
         if (megsb[eid[i].nli] && mcb[m]) {
	    eg[eid[i].nli][eid[i].ci].w = 1;
	    trainEg.push_back(eg[eid[i].nli][eid[i].ci]);
	    e[eid[i].nli].push_back(m);
	    megsb[eid[i].nli] = false;
	    mcb[m] = false;
         }
      }
   }*/

   // uses maximum weight assignment
   vector <vector <double> > hung(megs.size(), vector <double>(mc.size(),0));
   for (int i=0; i<eid.size(); ++i) {
      int m = megs[eid[i].nli].mci[eid[i].ci];
      if (eid[i].conf >= THR) hung[eid[i].nli][m] = eid[i].conf+100000;
   }
   //cout<<"here 3\n";
   vector <int> solu;
   solve_hungarian(hung,solu);
   //cout<<"here 4\n";
   for (int i=0; i<megs.size(); ++i) {
      if (OUT) cout<<"-------------------------------\n"<<megs[i]<<"\n";
      int a=-1;
      //if (hung[i][solu[i]] < 0) {
	// cout<<"Not matched\n";
	 //continue;
      //}
      for (int j=0; j<megs[i].mci.size(); ++j) {
	 if (solu[i]==megs[i].mci[j] && hung[i][solu[i]] > 0) {
	    a=j;
	    break;
	 }
      }
      for (int j=0; j<megs[i].mci.size(); ++j) {
	 if (OUT) {
	    if (a==j) cout<<"--> ";
	    else cout<<"    ";
	    cout<<eg[i][j].c<<" ";
	    if (eg[i][j].ans.size() > 0)
	       cout<<"conf="<<eg[i][j].ans[0].conf<<"\n";
	    else cout<<" could not parse\n";
	 }
      }
      //assert(a != -1); 
      if (a != -1) {
         //cout<<solu[i]<<" "<<a<<" "<<megs[i].mci.size()<<"\n";
         eg[i][a].w = 1;
         trainEg.push_back(eg[i][a]);
         e[i].push_back(solu[i]);
      }
      else {
	 if (OUT) cout<<"Not matched\n";
      }
      //cout<<i<<"-\n";
   }
   
   //cout<<"here 5\n";
   replaceConstWords(trainEg);
   //cout<<"here 6\n";
}

void generateWeightedEgs(const vector <multEg> &megs, const semanticParser &sp, vector <example> &trainEg, vector <vector <int> > &e) {
   e.clear();
   e.resize(megs.size());
   vector <vector <example> > eg(megs.size());
   vector <egId> eid;
   for (int i=0; i<megs.size(); ++i) {
      eg[i].resize(megs[i].mci.size());
      for (int j=0; j<megs[i].mci.size(); ++j) {
	 eg[i][j] = makeExample(megs[i].nl, mc[megs[i].mci[j]]);
	 eid.push_back(egId(i,j));
	 vector <bool> rec(prods.size(), true);
         for (int p=0; p<prods.size(); ++p) {
	    rec[p] = true;
	    if (eg[i][j].y[p] <= 0) {
	       rec[p] = false;
	    }
         }
	 sp.earleyParses(START, eg[i][j], rec, 1, 0, true);
	 if (eg[i][j].ans.size() > 0) eid[eid.size()-1].conf = eg[i][j].ans[0].conf;
      }
   }

   for (int itr=0; itr<10; ++itr) {
      vector <double> n(megs.size(), 0); 
      for (int j=0; j<eid.size(); ++j) n[eid[j].nli] += eid[j].conf;
      for (int j=0; j<eid.size(); ++j) {
	 if (n[eid[j].nli] != 0) eid[j].conf /= n[eid[j].nli];
      }
      vector <double> c(mc.size(), 0); 
      for (int j=0; j<eid.size(); ++j) n[eid[j].ci] += eid[j].conf;
      for (int j=0; j<eid.size(); ++j) {
	 if (c[eid[j].ci] != 0) eid[j].conf /= c[eid[j].ci];
      }
   }
   
   vector <double> max(megs.size(), 0);
   vector <int> maxm(megs.size(), -1);
   for (int i=0; i<eid.size(); ++i) {
      if (eid[i].conf > 0) {
         int m = megs[eid[i].nli].mci[eid[i].ci];
	 eg[eid[i].nli][eid[i].ci].w = eid[i].conf;
	 trainEg.push_back(eg[eid[i].nli][eid[i].ci]);
	 if (max[eid[i].nli] < eid[i].conf) {
	    max[eid[i].nli] = eid[i].conf;
	    maxm[eid[i].nli] = m;
	 }
      }
   }
   for (int i=0; i<maxm.size(); ++i) { // highest used for convergence criteria
      if (maxm[i] != -1) {
	 e[i].push_back(maxm[i]);
      }
   }
   replaceConstWords(trainEg);
}


void getConnectedSubgraphs(const vector <vector <double> > &hung, vector <vector <int> > &r, vector <vector <int> > &c) {

   int a=0;
   while (a < hung.size()) {
      vector <int> ar, ac;
      ar.push_back(a);

      int i=0;
      while ( i < ar.size()) {
	 a = ar[i];
	 for (int j=0; j<hung[a].size(); ++j) {
	    if (hung[a][j] != 0) {
	       if (find(ac.begin(),ac.end(),j) == ac.end()) {
		  ac.push_back(j);
		  for (int e=0; e<hung.size(); ++e) {
		     if (hung[e][j] != 0) {
			if (find(ar.begin(),ar.end(),e) == ar.end()) {
			   ar.push_back(e);
			}
		     }
		  }
	       }
	    }
	 }
	 ++i;
	 if (i < ar.size()) a = ar[i];
      }


      sort(ar.begin(),ar.end());
      sort(ac.begin(),ac.end());

      a = ar[ar.size()-1]+1;

      r.push_back(ar);
      c.push_back(ac);
   }
}

void reduceAmbiguity(vector <multEg> &megs) {
   // uses matching algorithm to remove easily removable ambiguities
   int total=0, removed=0;
   vector <vector <double> > hung(megs.size(), vector <double>(mc.size(),0));
   for (int i=0; i<megs.size(); ++i) {
      for (int j=0; j<megs[i].mci.size(); ++j) {
	 hung[i][megs[i].mci[j]]=1;
	 ++total;
      }
   }

   vector <vector <int> > r, c;
   getConnectedSubgraphs(hung,r,c);
   for (int a=0; a<r.size(); ++a) {
      vector <vector <double> > h(r[a].size(), vector <double> (c[a].size(),0));
      for (int i=0; i<r[a].size(); ++i) {
         for (int j=0; j<c[a].size(); ++j) {
	    h[i][j] = hung[r[a][i]][c[a][j]];
	 }
      }
      bool change=true;
      while (change) {
         change=false;
         vector <double> storeMR(r[a].size());
         vector <double> storeS(c[a].size());
	 
         for (int i=0; i<r[a].size(); ++i) {
	    //cout<<"sentence: "<<r[a][i]<<"\n";
	    int j=0;
            while (j<c[a].size()) {
	       if (h[i][j] == 0) {
		  ++j;
		  continue;
	       }
               for (int jj=0; jj<c[a].size(); ++jj) { 
	          // temporarily delete edges with this NL
	          storeS[jj]=h[i][jj];
	          h[i][jj]=0;
	       }
	       for (int ii=0; ii<r[a].size(); ++ii) { 
	          // temporarily delete edges with this MR (but store them)
	          storeMR[ii] = h[ii][j];
	          h[ii][j]=0;
	       }
	       vector <int> s;
	       double w = solve_hungarian(h,s);
	       if (w < h.size()-1) {
	          // matching not complete, remove this NL-MR pair
		  int er=0;
		  while (er < megs[r[a][i]].mci.size()) {
		     if (megs[r[a][i]].mci[er]==c[a][j]) {
			megs[r[a][i]].mci.erase(megs[r[a][i]].mci.begin()+er);
			break;
		     }
		     else ++er;
		  }
	          change=true;
	          //cout<<"Removing a pair\n";
		  ++removed;
	          storeS[j]=0;
	          storeMR[i]=0;
	       }
	       for (int ii=0; ii<h.size(); ++ii) { 
	          // put the edges with that MR back
	          h[ii][j]=storeMR[ii];
	       }
	       for (int jj=0; jj<c[a].size(); ++jj) { 
	          // put the edges back with that NL back
	          h[i][jj]=storeS[jj];
	       }
	       ++j;
	    }
         }
      }
   }
   cout<<"Initial ambiguities:\n";
   cout<<"  Edges removed: "<<removed<<" Total: "<<total<<" "<<(100.0*removed)/total<<" %\n";
}

void mult_train_test(vector <multEg> &megs, vector <example> &testEg) {
   vector <example> trainEg;

   reduceAmbiguity(megs);

   semanticParser spFinal;
   ANN=1;

   int a=0;
   vector <multEg> m;
   vector <vector <int> > ep;
   semanticParser prev;
   for (int i=0; i<megs.size(); ++i) {
      m.push_back(megs[i]);
      a += megs[i].mci.size();
      if (a > MAX_FIRST_ITR_EGS || i==megs.size()-1) {
	 vector <vector <int> > ee;
	 vector <example> eg;
	 generateEgs(m,eg,ee);
	 if (eg.size() >= 25) {
	    semanticParser sp;
	    //cout<<eg.size()<<" ---\n";
	    sp.train(eg,prods);
	    prev=sp;
	    generateBestEgs(m,sp,trainEg,ee,THR1);
	 }
	 else {
	    generateBestEgs(m,prev,trainEg,ee,THR1);
	 }
	 m.clear();
	 for (int x=0; x<ee.size(); ++x) ep.push_back(ee[x]);
	 a=0;
      }
   }
   cout<<"Initial iteration over\n";

   vector <vector <int> > e;
   for (int itr=0; itr<MAX_ITR; ++itr) {
      cout<<"Iteration: "<<itr<<"\n";
      semanticParser sp;
      sp.train(trainEg, prods);
      trainEg.clear();
      cout<<"In..\n";
      generateBestEgs(megs, sp, trainEg,e,THR2);
      cout<<"out\n";

      // see if it has converged
      double c=0;
      bool converged=false;
      for (int i=0; i<e.size(); ++i) {
	 if (e[i]==ep[i]) ++c;
      }
      cout<<"\nConverged: "<<100*c/e.size()<<" %\n\n";
      if (c/e.size() > CONVG) converged=true;
      else ep = e;
      
      if (itr==MAX_ITR-1 || converged) {
	 spFinal = sp;
	 break;
      }
      else {
         sp.test(testEg);
         //string s = itos(itr);
         //writeOut(testEg, string(outdir+"itr"+s+"-test-out"));
         ANN = (int)power(5,itr);
      }
   }
   spFinal.test(testEg);
}

/*
void mult_train_test(vector <multEg> &megs, vector <example> &testEg) {
   int MAX_ITR=5;
   double CONVG=0.8; // convergence => 80% same in the next iteration
   vector <example> trainEg;

   reduceAmbiguity(megs);

   semanticParser spFinal;
   vector <vector <int> > e, ep;
   generateEgs(megs, trainEg,ep);
   ANN=1;
   for (int itr=0; itr<MAX_ITR; ++itr) {
      cout<<"Iteration: "<<itr<<"\n";
      semanticParser sp;
      sp.train(trainEg, prods);
      trainEg.clear();
      cout<<"In..\n";
      generateBestEgs(megs, sp, trainEg,e);
      //generateWeightedEgs(megs, sp, trainEg,e);
      cout<<"out\n";

      // see if it has converged
      double c=0;
      bool converged=false;
      for (int i=0; i<e.size(); ++i) {
	 if (e[i]==ep[i]) ++c;
      }
      cout<<c/e.size()<<" =========\n";
      if (c/e.size() > CONVG) converged=true;
      else ep = e;
      
      if (itr==MAX_ITR-1 || converged) {
	 spFinal = sp;
	 break;
      }
      else {
         sp.test(testEg);
         string s = itos(itr);
         writeOut(testEg, string(outdir+"itr"+s+"-test-out"));
         ANN = (int)power(5,itr);
      }
   }
   spFinal.test(testEg);
}*/
