#include <cassert>
#include <map>

#include "production.h"

extern map <string, int> s2n; // mapping from strings to numbers
extern vector <string> n2s; // reverse mapping

production::production(const string &l, const vector <string> &r, int lev) {
   computedNt=false;
   lhs = l;
   rhs = r;
   nt.resize(rhs.size());
   for (int i=0; i<nt.size(); ++i) nt[i] = ifNt(rhs[i]);
   computeNt();
}


bool production::read(ifstream &in) {
   // reads a grammar rule from the file
   // first reads and forgets up to "<prod>" 

   string w;
   
   while ((w != "<prod>") && in){
      in>>w;
   }

   if (! in) return false;

   /*
   in>>w;
   string sn = "";
   for (int i=7; i<w.length()-2; ++i) {
      sn.push_back(w[i]);
   }
   level = atoi(sn.c_str());*/

   in>>w;
   assert(w[0]=='*');

   lhs = w;
   bool endf = false;
         
   in>>w;
   assert(w=="->");
   
   while (in) {
         in>>w;
         if (w=="</prod>" || w=="<pat>") {
            endf = true;
            break;
	 }
	 rhs.push_back(w);
         if (w[0]=='*') { // non-terminal
	    nt.push_back(true);
         }
         else { // terminal
	    nt.push_back(false);
         }
   }
   assert(endf);

   while (w=="<pat>") {
      in>>w;
      vector <string> s;
      vector <int> si;
      while (w != "</pat>") {
	 s.push_back(w);
	 map <string, int>::iterator itr=s2n.find(w);
	 if (itr==s2n.end()) {
	    s2n[w] = n2s.size();
	    si.push_back(n2s.size());
	    n2s.push_back(w);
	 }
	 else {
	    si.push_back(itr->second);
	 }
         in>>w;
      }
      if (s.size() > 0) {
	 pat.push_back(s);
	 pati.push_back(si);
      }
      in>>w;
   }
   computeNt();
   
   return true;
}

bool production::patPresent(const vector <string> &s, int start, int end) const {
   // true if any pattern is present in s
   for (int i=0; i<pat.size(); ++i) {
      for (int j=start; j<=int(s.size()-pat[i].size()) && j<=int(end-pat[i].size()+1); ++j) { 
	    bool r = true;
	    for (int p=0; p<pat[i].size(); ++p) {
	       if (pat[i][p] != s[j+p]) {
		  r = false;
		  break;
	       }
	    }
	    if (r) return true;
	 }
   }
   return false;
}

bool production::patPresent(const vector <int> &s, int start, int end) const {
   // true if any pattern is present in s
   for (int i=0; i<pati.size(); ++i) {
      for (int j=start; j<=int(s.size()-pat[i].size()) && j<=int(end-pat[i].size()+1); ++j) { 
	    bool r = true;
	    for (int p=0; p<pati[i].size(); ++p) {
	       if (pati[i][p] != s[j+p]) {
		  r = false;
		  break;
	       }
	    }
	    if (r) return true;
	 }
   }
   return false;
}

ostream& production::write(ostream &out) const {
     out<<"<prod> "<<lhs<<" -> ";
     for (int i=0; i<rhs.size(); ++i) out<<rhs[i]<<" ";
     out<<"</prod>\n";
     return out;
}
