#ifndef _PRODUCTION_H
#define _PRODUCTION_H

#include <string>
#include <vector>
#include <fstream>

using namespace std;

class production {
   string lhs;
   vector <string> rhs;
   vector <bool> nt; // if the rhs[i] is a non-terminal

   bool computedNt; // internal, true if rhsNt has been computed
   vector <string> rhsNt;

   void computeNt() {
      if (! computedNt) {
         for (int i=0; i<rhs.size(); ++i) {
	    if (ifNt(rhs[i])) rhsNt.push_back(rhs[i]);
         }
         computedNt = true;
      }
      if (rhsNt.size()==0) level = 0;
      else level = 1;
   }

   
   public:
   int level;
   int pn; // production number
   bool isConst, learnable, dupl;
   vector <vector <string> > pat; // given patterns 
   vector <vector <int> > pati; // mapped ints of given patterns 
   
   static bool ifNt(const string &s) {
      return (s.size() > 0 && s[0]=='*');
   }

   production() {computedNt=false; level=0;}
   
   production(const string &l, const vector <string> &r, int lev=0);

   bool operator==(const production &other) const {
      return lhs==other.lhs && rhs==other.rhs;
   }

/*   bool operator < (const production &other) const {
      int m=0, om=0;
      for (int i=0; i<pat.size(); ++i) {
	 if (pat[i].size() > m) m = pat[i].size();
      }
      for (int i=0; i<other.pat.size(); ++i) {
	 if (other.pat[i].size() > om) om = other.pat[i].size();
      }
      return m > om;
   }*/
   
   const string &getLhs() const {
      return lhs;
   }
   const vector <string> &getRhs() const {
      return rhs;
   }
   void putRhs(const vector <string> &prhs) {
      rhs = prhs;
      rhsNt.clear();
      computedNt=false;
      nt.resize(rhs.size());
      for (int i=0; i<nt.size(); ++i) {
	 nt[i] = ifNt(rhs[i]);
	 if (level==0 && nt[i]) ++level; 
      }
      computeNt();
   }
   vector <string> getNt() const {
      return rhsNt;
   }
   string giveNti(int i) const {
      return rhsNt[i];
   }
   int ntSize() const {
      return rhsNt.size();
   }
   bool read(ifstream &in);

   bool patPresent(const vector <string> &nl, int s=0, int e=999999) const;
   bool patPresent(const vector <int> &onli, int s=0, int e=999999) const;

   ostream& write(ostream &out) const;

};

inline ostream& operator << (ostream& out, const  production& p) {
       return p.write(out);
}
#endif
