#include "semanticParser.h"

extern string START;
extern int ITR;
extern double LIM;
extern int BEAM;
extern int MAX_MEM;
extern vector <vector <int> > nli;
extern map<string, int> nonterm;

extern bool DEBUG;

extern int ANN;

bool ALLOW=false;

bool RECORD=false;
vector <vector <int> > q;

void semanticParser::collectPosEg(const clang &c, int ind) {
   // collect the portions as positive

   example &eg = trainEg[ind];
   
   if (prods[c.rootProd].learnable) {

      vector <int> n;
      assert(c.pe < eg.nli.size());
      assert(c.ps <= c.pe);
      assert(c.ps >= 0);
      for (int j=c.ps;  j<=c.pe; ++j) { 
	 n.push_back(eg.nli[j]);
      }
      int f=0;
      for (f=0; f<nli.size(); ++f) {
	 if (nli[f]==n) break;
      }
      if (f < nli.size()) {
	// if (find(pos[c.rootProd].begin(), pos[c.rootProd].end(), pn(f))==pos[c.rootProd].end()) {
	    pos[c.rootProd].push_back(pn(f, eg.w));
	    //if (eg.w < 1) {
	      // neg[c.rootProd].push_back(pn(f, 1-eg.w));
	    //}
	 //}
      }
      else {
	 nli.push_back(n);
	 pos[c.rootProd].push_back(pn(nli.size()-1, eg.w));
	 //if (eg.w < 1) {
	   // neg[c.rootProd].push_back(pn(nli.size()-1, 1-eg.w));
	 //}
      }
   }
   // recursive
   for (int i=0; i<c.childSize(); ++i) {
      if (c.child[i] != NULL && production::ifNt(c.child[i]->s)) {
	 collectPosEg(*c.child[i], ind);
      }
   }
}

 // depth-first traversal
bool semanticParser::collectNegHelper(clang &c, const clang &corr, vector <bool> &prob) const {
   // finds out what portions of the sentence are incorrectly covered
   if (c.rootProd != corr.rootProd && prods[c.rootProd].learnable) {
      for (int j=c.ps;  j<=c.pe; ++j) {
	 prob[j] = true;
      }
      for (int j=corr.ps;  j<=corr.pe; ++j) {
	 prob[j] = true;
      }
      return true;
   }
   for (int i=0; i<c.childSize(); ++i) {
      if (c.child[i] != NULL) {
	 if (production::ifNt(c.child[i]->s)) {
	    if (prods[c.child[i]->rootProd].learnable) {
	       if (collectNegHelper(*c.child[i], *corr.child[i], prob)) return true;
	    }
	    else { 
	       if (c.child[i]->rootProd != corr.child[i]->rootProd) {
		  // if consts differ then the portions covered by parents 
		  // are taken as incorrect
		  for (int j=c.ps;  j<=c.pe; ++j) {
		     prob[j] = true;
		  }
		  for (int j=corr.ps;  j<=corr.pe; ++j) {
		     prob[j] = true;
		  }
		  return true;
	       }
	    }
	 }
      }
   }
   return false;
}


bool semanticParser::bfs(vector <clang*> c, vector <clang*> corr, int q, vector <bool> &prob) const {
   if (q==c.size()) return false;

   if (c[q]->rootProd != corr[q]->rootProd && prods[c[q]->rootProd].learnable) {
      for (int j=c[q]->ps;  j<=c[q]->pe; ++j) {
	 prob[j] = true;
      }
      for (int j=corr[q]->ps;  j<=corr[q]->pe; ++j) {
	 prob[j] = true;
      }
      return true;
   }

   for (int i=0; i<c[q]->childSize(); ++i) {
      if (c[q]->child[i] != NULL) {
	 if (production::ifNt(c[q]->child[i]->s)) {
	    if (prods[c[q]->child[i]->rootProd].learnable) {
	       c.push_back(c[q]->child[i]);
	       corr.push_back(corr[q]->child[i]);
	    }
	    else { 
	       if (c[q]->child[i]->rootProd != corr[q]->child[i]->rootProd) {
		  // if consts differ then the portions covered by parents 
		  // are taken as incorrect
		  for (int j=c[q]->ps;  j<=c[q]->pe; ++j) {
		     prob[j] = true;
		  }
		  for (int j=corr[q]->ps;  j<=corr[q]->pe; ++j) {
		     prob[j] = true;
		  }
		  return true;
	       }
	    }
	 }
      }
   }
   return bfs(c,corr,q+1,prob);
}

void semanticParser::wordCovered(int w, const clang &c, vector <int> &cov, vector <int> &ids)  const {
   // finds productions cover the wth word and their start and end points
   if (w >= c.ps && w<= c.pe) {
      cov.push_back(c.rootProd);
      ids.push_back(c.nodeId);
   }
   for (int i=0; i<c.childSize(); ++i) {
      if (c.child[i] != NULL && production::ifNt(c.child[i]->s)) {
	 wordCovered(w, *c.child[i], cov, ids);
      }
   }
}

void semanticParser::wordCovered(int w, const clang &c, vector <int> &cov) const {
   // what productions cover the wth word
   if (w >= c.ps && w<= c.pe) {
      cov.push_back(c.rootProd);
   }
   for (int i=0; i<c.childSize(); ++i) {
      if (c.child[i] != NULL && production::ifNt(c.child[i]->s)) {
	 wordCovered(w, *c.child[i], cov);
      }
   }
}
 
void semanticParser::getNegs(const clang &c, vector <nega> &ngs) const {
   if (c.neg) {

      nega ng(c.rootProd, c.ps, c.pe);
      if (find(ngs.begin(), ngs.end(), ng)==ngs.end()) {
	 ngs.push_back(ng);
      }
   }
   for (int i=0; i<c.childSize(); ++i) {
      if (c.child[i] != NULL && production::ifNt(c.child[i]->s)) {
	 getNegs(*c.child[i], ngs);
      }
   }
}

void semanticParser::collectNegFromPos() {
   for (int p=0; p<pos.size(); ++p) {
         for (int pp=0; pp<prods.size(); ++pp) {
	    if (pp != p && /*prods[pp].level==0 &&*/ prods[pp].getLhs()==prods[p].getLhs()) {
	       for (int i=0; i<pos[p].size(); ++i) {
		  if (find(neg[pp].begin(), neg[pp].end(), pn(pos[p][i]))==neg[pp].end()) {
		     neg[pp].push_back(pos[p][i]);
		  }
	       }
	    }
	 }
   }
}

void semanticParser::collectNegEg(clang &c, const clang &corr, int i) {
   example &eg =  trainEg[i];
   vector <bool> prob(eg.nli.size(), false);
   if (collectNegHelper(c, corr, prob)) {
      c.giveNodeIds();
      vector <int> negIds;
      vector <nega> ngs;
      for (int i=0; i<prob.size(); ++i) {
	 if (prob[i]) {
	    vector <int> ccov, corrcov, ids; 
	    wordCovered(i,c,ccov, ids);
	    wordCovered(i,corr,corrcov);
	    for (int j=0; j<ccov.size(); ++j) {
	       if (find(corrcov.begin(), corrcov.end(), ccov[j])==corrcov.end()) {
		  // a production that covered a word did not cover it in the correct derivation => negative example
		  negIds.push_back(ids[j]);
	       }
	    }
	 }
      }
      c.markNegs(negIds);

      getNegs(c, ngs);
      
      // convert the neg examples
      for (int i=0; i<ngs.size(); ++i) {
	 vector <int> n; 
         for (int j=ngs[i].s;  j<=ngs[i].e; ++j) {
	    n.push_back(eg.nli[j]);
	 }
         int f=0;
	 for (f=0; f<nli.size(); ++f) {
	    if (nli[f]==n) break;
	 }
         if (f < nli.size()) {
	  //  if (find(neg[ngs[i].p].begin(), neg[ngs[i].p].end(), pn(f))==neg[ngs[i].p].end()) {
	       neg[ngs[i].p].push_back(pn(f, eg.w));
	    //}
         }
         else {
	    nli.push_back(n);
	    neg[ngs[i].p].push_back(pn(nli.size()-1, eg.w));
         }
      }
   }
}

void semanticParser::collectPosNegEg(int i) {

   example &eg = trainEg[i];

   int corr=-1;
   double maxC=-1;

   //cout<<"-----------------------------------\n";
   //cout<<eg<<"\n";
   
   for (int a=0; a<eg.ans.size(); ++a) { // find the max conf correct parse
      if (find(eg.mc.begin(), eg.mc.end(), eg.ans[a]) != eg.mc.end()) { 
	 // matches an MR
	 if (eg.ans[a].conf > maxC) {
	    maxC = eg.ans[a].conf;
	    corr=a;
	 }
      }
   }
   if (corr != -1) {
         // collect positives
	    
         collectPosEg(eg.ans[corr], i);
         for (int a=0; a<eg.ans.size(); ++a) {
	    if (a==corr) {
              // cout<<eg.ans[a]<<eg.ans[a].conf<<"\n";
	      // cout<<"^^^^^^^\n\n";
	       break;
	    }
	 }
         for (int a=0; a<eg.ans.size(); ++a) {
            //cout<<eg.ans[a]<<eg.ans[a].conf<<"\n";
            if (a != corr && eg.ans[a].conf > eg.ans[corr].conf) {
	    // collect negatives
	       //cout<<"NNNNNNN\n\n";
	       collectNegEg(eg.ans[a], eg.ans[corr], i);
	       //break;
            }
         }
   }
}


bool semanticParser::constantString (const string &s) const {
#ifndef _CLANG
      if (s=="*cityid" || s=="*placeid" || s=="*riverid" ||s=="*countryidid" ||s=="*stateid") {
#endif
#ifdef _CLANG
      if (s=="*NUM" || s=="*CLANG_STR") {
#endif
	 return true;
      }
      else return false;
}
double semanticParser::findProb(int p, int s, int e, example &teg) const {
   return sc[p].giveProb(s,e,teg);

}
double semanticParser::findProb(int p, int s, int e, example &teg, double &prob) const {
   if (prob==-1) {
      if (RECORD) {
	 q[p].push_back(s);
	 q[p].push_back(e);
      }
      if (scTrained) {
	 prob = sc[p].giveProb(s,e,teg);
      }
      else {
	 if (initScTrained) {
	    prob = initSc[p].giveProb(s,e,teg);
	 }
         else {
	    cerr<<"Nothing trained\n";
	    exit(1);
         }
      }
   }
   return prob;
}

void semanticParser::scanner(int w, vector <vector <earleyState> > &chart, int end, int ind, example &teg, vector <vector <vector <double> > > &probs, const vector <bool> &rec, bool force) const {
   if (rec[chart[end][ind].p]) {
      double conf = findProb(chart[end][ind].p, chart[end][ind].end, w, teg, probs[chart[end][ind].p][chart[end][ind].end][w]);
      if ((force/* && conf >=0*/) || (!force && conf > LIM)) {
	 earleyState n(chart[end][ind].begin, w+1, 1, chart[end][ind].p, conf);
	 //if (DEBUG) cout<<"From scanner: "<<n<<"\n";
	 assert(n.prob <= 1);
	 chart[n.end].push_back(n);
      }
   }
}

void semanticParser::predictor(int w, vector <vector <earleyState> > &chart, int end, int ind, example &teg, vector <vector <vector <double> > > &probs, const vector <bool> &rec, bool force) const {
   string nt = chart[end][ind].rhs[chart[end][ind].dot];
   //cout<<"Predicting using: "<<chart[end][ind]<<"\n";
   for (int p=0; p<prods.size(); ++p) {
      if (rec[p] && nt==prods[p].getLhs()) {
	 if (prods[p].level==0) {
	    earleyState n(w,w,0,p,1);
	    if (find(chart[w].begin(), chart[w].end(), n) == chart[w].end()) {
	       bool in = false;
	       // don't insert if the production has no good probability 
	       // anywhere later
	       if (!force) {
	          for (int i=n.begin; i<teg.nli.size(); ++i) {
	             if (findProb(n.p,n.begin,i,teg,probs[n.p][n.begin][i]) > LIM) {
		        in = true;
		        break;
	            }
	          }
	       }
	       else {
		  in=true;
	       }
	       if (in) {
	          //if (DEBUG) cout<<"From predictor: "<<n<<"\n";
	          //assert(n.prob <= 1);
	          chart[w].push_back(n);
	       }
	    }
	 }
	 else {
	    vector <string> rhs = prods[p].getNt();
	    sort(rhs.begin(), rhs.end());
	    bool np=false;
	    // don't insert if the production has no good probability 
	    // anywhere later
	    if (!force) {
	       for (int i=w; i<teg.nli.size(); ++i) {
	          if (findProb(p,w,i,teg,probs[p][w][i]) > LIM) {
		     np = true;
		     break;
	          }
	       }
	    }
	    else {
	       np=true;
	    }
	    while (np) {  // insert all the permutations
	       earleyState n = earleyState(w, w, 0, p, rhs, 1);
	       if (find(chart[w].begin(), chart[w].end(), n) == chart[w].end()) {
	          //if (DEBUG) cout<<"From predictor: "<<n<<"\n";
	 //assert(n.prob <= 1);
	          chart[w].push_back(n);
	       }
	       np = false;
	       np = next_permutation(rhs.begin(), rhs.end());
	    }
	 }
      }
   }
}

void fixComingProbs(vector <vector <earleyState> > &chart, const vector <earleyStateIndex> &ei, int a, int b) {
   for (int i=0; i<ei.size(); ++i) {
      for (int k=0; k<chart[ei[i].end][ei[i].ind].going.size(); ++k) {
	 if (chart[ei[i].end][ei[i].ind].going[k].end == a && chart[ei[i].end][ei[i].ind].going[k].ind==b) {
	    chart[ei[i].end][ei[i].ind].going.erase(chart[ei[i].end][ei[i].ind].going.begin()+k);
	    break;
	 }
      }
   }
}

void fixGoingProbs(vector <vector <earleyState> > &chart, const vector <earleyStateIndex> &ei, int a, int b, double nprob, double oprob) {
   for (int i=0; i<ei.size(); ++i) {
      chart[ei[i].end][ei[i].ind].prob /= oprob; 
      chart[ei[i].end][ei[i].ind].prob *= nprob; 
      fixGoingProbs(chart, chart[ei[i].end][ei[i].ind].going, ei[i].end, ei[i].ind, nprob, oprob);
   }
}


void semanticParser::completer(int w, vector <vector <earleyState> > &chart, int end, int ind, example &teg, vector <vector <vector <double> > > &probs, const vector <bool> &rec, int beam, double corrLim, bool force) const {
   //cout<<"In completer: "<<chart[end][ind]<<"\n\n";
   int nt = nonterm[chart[end][ind].lhs];
   int i=0;
   while (i<chart[chart[end][ind].begin].size()) {
      const earleyState &es = chart[chart[end][ind].begin][i];
      if (es.complete() || nonterm[es.rhs[es.dot]] != nt) {
	 ++i;
	 continue;
      }
      earleyState n = es;
      n.end = chart[end][ind].end;
      n.dot += 1;
      n.prob *= chart[end][ind].prob;
      n.coming.push_back(earleyStateIndex(end, ind));

      if (! n.complete()) {
         //if (find(chart[chart[end][ind].end].begin(), chart[chart[end][ind].end].end(), n) == chart[chart[end][ind].end].end()) {
	    //if (DEBUG) cout<<"From completer: \n"<<es<<n<<"\n";
	    chart[n.end].push_back(n);
	    chart[end][ind].going.push_back(earleyStateIndex(n.end, chart[n.end].size()-1));
         //}
      }
      else {
	 if (force) {
	    clang tmp = recover(n,chart);
	    //if (DEBUG) 
	    //cout<<tmp<<"\n\n";
	    if (! teg.c.subTreeExists(tmp)) {
	       ++i;
	       continue;
	    }
	    //if (DEBUG) 
	    //cout<<"Included\n\n\n";
	 }
	 double conf;
	 if (n.p==-1) conf=1;
	 else conf = findProb(n.p,n.begin,n.end-1,teg,probs[n.p][n.begin][n.end-1]);
	 if ((force/* && conf >=0*/) || (!force && conf > LIM)) {
	    //if (DEBUG) cout<<"From completer: \n"<<es<<n<<"\n";
	    bool in=true;
	    n.prob *= conf;
	    int ex=0;
	    int minj=-1;
	    double min=9999;
	    for (int j=0; j<chart[n.end].size(); ++j) {
	       if (chart[n.end][j]==n) {
		  /*
		  if (chart[n.end][j].prob < n.prob) {
		     double op = chart[n.end][j].prob;
		     n.going = chart[n.end][j].going;
		     chart[n.end][j] = n;
		     fixGoingProbs(chart, n.going, n.prob, op);
		  }
		  in=false;
		  break;*/
		  ++ex;
		  if (chart[n.end][j].prob < min) {
		     min = chart[n.end][j].prob;
		     minj=j;
		  }
	       }
	    }
	    if (ex >= beam) {
	       // beam is full, replace the lowest one if it is lower than the new prob
	       if (n.prob > min && minj >= 0) {
		  fixComingProbs(chart, chart[n.end][minj].coming, n.end, minj);
		  n.going = chart[n.end][minj].going;
		  chart[n.end][minj] = n;
	          chart[end][ind].going.push_back(earleyStateIndex(n.end,minj));
		  fixGoingProbs(chart, n.going, n.end, minj, n.prob, min);
	       }
	       in=false;
	    }
	    if (in) {
	       chart[n.end].push_back(n);
	       chart[end][ind].going.push_back(earleyStateIndex(n.end,chart[n.end].size()-1));
	    }
	 }
      }
      ++i;
   }
}

clang semanticParser::recover(const earleyState &es, const vector <vector <earleyState> > &chart, int depth) const {
   if (es.p==-1) {
      assert(es.coming.size() > 0);
      return recover(chart[es.coming[0].end][es.coming[0].ind], chart,depth+1);
   }
   clang r;
   r.s = es.lhs;
   r.conf = es.prob;
   r.selfConf = es.selfProb;
   r.ps = es.begin;
   r.pe = es.end-1;
   r.rootProd = es.p;
   vector <bool> re(es.coming.size(), false);
   const vector <string> &rhs = prods[es.p].getRhs();
   for (int i=0; i<rhs.size(); ++i) {
      if (production::ifNt(rhs[i])) {
	 int c;
	 for (c=0; c<es.coming.size(); ++c) {
	    if (!re[c] && chart[es.coming[c].end][es.coming[c].ind].lhs==rhs[i]) {
	       re[c] = true;
	       break;
	    }
	 }
	 if (c >=  es.coming.size()) {
	    cout<<es<<"\n";
	    cout<<rhs[i]<<"\n";
	    for (int a=0; a<es.coming.size(); ++a) {
	       cout<<chart[es.coming[a].end][es.coming[a].ind]<<"\n";
	    }
	 }
	 assert(c<es.coming.size());
	 r.makeChild(recover(chart[es.coming[c].end][es.coming[c].ind], chart,depth+1),i);
      }
      else {
	 r.makeChild(clang(rhs[i]),i);
      }
   }
   return r;
}


void semanticParser::earleyParses_helper(const string &root, example &teg, const vector <bool> &rec, int beam, double corrLim, bool force) const {

   // store of probabilities, prob[p][start][end]
   vector <vector <vector <double> > > probs(prods.size(), vector <vector <double> > (teg.nli.size(), vector <double> (teg.nli.size(),-1)));

   vector <vector <earleyState> > chart(teg.nli.size()+1); 

   // insert null state
   chart[0].push_back(earleyState(root)); 

   for (int w=0; w<teg.nli.size(); ++w) {
      // process each word 
      bool cons = false;
      string consStr;
      if (constantString(teg.nl[w])) {
	 cons=true;
	 consStr = teg.nl[w];
      }
      // scan previous states first
      for (int end=0; end<w; ++end) {
	 int i=0;
	 while (i<chart[end].size()) {
	    earleyState &es = chart[end][i];
	    if (! es.complete()) {
	       if (es.p != -1 && prods[es.p].level==0) { // base state
	          if (!cons || (es.lhs == consStr && !prods[es.p].learnable)) {
		     scanner(w,chart,end,i,teg,probs,rec,force);
	          }
	       }
	    }
	    ++i;
	 }
      }
      int end = w;
      int i=0;
      while (i<chart[end].size()) {
	 earleyState &es = chart[end][i];
	 if (! es.complete()) {
	    if (es.p != -1 && prods[es.p].level==0) { // base state
	       if (!cons || (es.lhs == consStr && !prods[es.p].learnable)) {
		  scanner(w,chart,end,i,teg,probs,rec,force);
	       }
	    }
	    else {
	       predictor(w,chart,end,i,teg,probs,rec,force);
	    }
	 }
	 else {
	    completer(w,chart,end,i,teg,probs,rec,beam,corrLim,force);
	 }
	 ++i;
      }
   }

   int end = teg.nli.size();
   for (int i=0; i<chart[end].size(); ++i) {
      if (chart[end][i].complete())
	 completer(end,chart,end,i,teg,probs,rec,beam,corrLim,force);
   }

   //vector <earleyState> cc;
   // get the answer
   //teg.ans.clear();
   for (int i=0; i<chart[end].size(); ++i) {
      if (chart[end][i].p==-1 && chart[end][i].complete()) {
	 //cc.push_back(chart[end][i]);
	 teg.ans.push_back(recover(chart[end][i], chart, i));
	 //break;  // only one answer
      }
   }

   /*
   if (teg.ans.size() > 1) {
      if (teg.ans[0].conf < teg.ans[1].conf) {
      cout<<teg.ans.size()<<" ------\n";
      //cout<<cc.size()<<" ------\n";
      for (int i=0; i<teg.ans.size(); ++i) {
         cout<<teg.ans[i]<<" "<<teg.ans[i].conf<<"\n";
	// cout<<cc[i].p<<" "<<cc[i].begin<<" "<<cc[i].end<<"\n";
	// cout<<cc[i]<<"\n";
      }
      }
      //sort(teg.ans.begin(), teg.ans.end());
      //teg.ans.resize(1);
   }*/

      sort(teg.ans.begin(), teg.ans.end());

   /*
   cout<<teg<<"\n";
   for (int i=0; i<teg.ans.size(); ++i) {
      cout<<teg.ans[i]<<" "<<teg.ans[i].conf<<"\n";
   }*/
}

void semanticParser::earleyParses(const string &root, example &teg, const vector <bool> &rec, int beam, double corrLim, bool force) const {
   double OLIM = LIM;
   teg.initMap();
   for (int i=0; i<3; ++i) {
      earleyParses_helper(root,teg,rec,beam,corrLim,force);
      if (teg.ans.size()==0) LIM /= 2;
      else break;
   }
   teg.resetMap(); // ?? not necessary
   LIM = OLIM;
}

int semanticParser::testOnTrain() {
   int incorr=0;
   for (int i=0; i<trainEg.size(); ++i) {
      trainEg[i].ans.clear();
#ifdef _CLANG
      if (  
	    trainEg[i].skip  || trainEg[i].id==38||trainEg[i].id==152||trainEg[i].id==157||trainEg[i].id==392||trainEg[i].id==411||trainEg[i].id==461
	    ) {
	 continue;
      }
#endif
      earleyParses(START, trainEg[i], recTr[i],BEAM,0);
      
      /*
      cout<<trainEg[i]<<"\n";
      cout<<trainEg[i].ans.size()<<"\n";
      for (int a=0; a<trainEg[i].ans.size(); ++a) { // if correct parse was found
	 cout<<trainEg[i].ans[a]<<"\n";
      }*/
      
      for (int e=0; e<trainEg[i].mc.size(); ++e) {
            int corr = -1;
            for (int a=0; a<trainEg[i].ans.size(); ++a) { // if correct parse was found
               if (trainEg[i].ans[a]==trainEg[i].mc[e]) { 
	          corr=a;
	          break;
	       }
            }
	    if (trainEg[i].ans.size() > 0 && corr!=0) ++incorr; // CH
            if (corr == -1) { // force to find the correct parse
	       vector <bool> rec(prods.size(), true);
	       for (int p=0; p<prods.size(); ++p) {
   	          if (trainEg[i].y[p] <= 0) {
	             rec[p] = false;
	          }
	       }
	       trainEg[i].c = trainEg[i].mc[e];
	       earleyParses(START, trainEg[i], rec, 1,0,true);
	    }
      }
      /*
      if (trainEg[i].ans.size()==0) {
	 cout<<trainEg[i]<<"\n";
      }
      cout<<"\n";*/
   }
   return incorr;
}

void semanticParser::collectPosNeg() {
   for (int p=0; p<pos.size(); ++p) {
      pos[p].clear();
   }
   
   for (int i=0; i<trainEg.size(); ++i) {
#ifdef _CLANG
      if (trainEg[i].skip || trainEg[i].id==38||trainEg[i].id==152||trainEg[i].id==157||trainEg[i].id==392||trainEg[i].id==411||trainEg[i].id==461) continue;
#endif
      if (trainEg[i].ans.size() > 0) {
	 collectPosNegEg(i);
      }
   }
   collectNegFromPos(); // CH
}

void semanticParser::markRec(vector <vector <bool> > &rec, vector <example> &eg) const{

   rec.resize(eg.size(), vector <bool> (prods.size(), true));
   
   for (int i=0; i<eg.size(); ++i) {
#ifdef _CLANG
     if (eg[i].skip || eg[i].id==38||eg[i].id==152||eg[i].id==157||eg[i].id==392||eg[i].id==411||eg[i].id==461) continue;
#endif
     eg[i].initMap();
     for (int p=0; p<prods.size(); ++p) {
	double prob;
	prob = initSc[p].giveProb(eg[i]);
	if (prob < LIM) {
	   rec[i][p]=false;
	}
     }
     eg[i].resetMap();
   }
}

void semanticParser::getInitialPosNeg() {
   initPos.resize(prods.size());
   initNeg.resize(prods.size());
   for (int i=0; i<trainEg.size(); ++i) {
      #ifdef _CLANG
      if (trainEg[i].skip || trainEg[i].id==38||trainEg[i].id==152||trainEg[i].id==157||trainEg[i].id==392||trainEg[i].id==411||trainEg[i].id==461) continue;
      #endif
      
      int f=0;
      for (f=0; f<nli.size(); ++f) {
	 if (nli[f]==trainEg[i].nli) break;
      }
      if (f==nli.size()) {
         nli.push_back(trainEg[i].nli);
      }
      //trainEg[i].c.writeParse(cout);
      for (int p=0; p<prods.size(); ++p) {
	 if (prods[p].learnable) {
	     if (trainEg[i].y[p]==1) {
		initPos[p].push_back(pn(f, trainEg[i].w));
		//if (trainEg[i].w < 1) {
		  // initNeg[p].push_back(pn(nli.size()-1, 1-trainEg[i].w));
		//}
	     }
	     else {
		initNeg[p].push_back(pn(f, trainEg[i].w));
	     }
	 }
      }
   }
}

void semanticParser::trainInitClassifiers() {
   //cout<<"Training init classifiers...\n";
   initSc.clear();
   initSc.resize(prods.size());

   for (int p=0; p<prods.size(); ++p) {
      if (initPos[p].size() > 0) assert(initPos[p][0].i>=0);
      /*
      if (initPos[p].size() > 0) {
         cout<<prods[p]<<"\n";
	 for (int j=0; j<initPos[p].size(); ++j) {
	    for (int i=0; i<nli[initPos[p][j].i].size(); ++i) {
	       cout<<n2s[nli[initPos[p][j].i][i]]<<" ";
	    }
	    cout<<"\n";
	 }
      }*/
      //if (initPos.size() > 62) assert(initPos[62][0].i >= 0);
      //cout<<prods[p]<<" --\n";
      initSc[p].train(initPos[p], initNeg[p], prods[p]);
      //if (initPos.size() > 62) assert(initPos[62][0].i >= 0);
   }
   initScTrained=true;
   //cout<<"Init classifiers trained.\n";
}

void semanticParser::trainClassifiers() {
   //cout<<"Training classifiers...\n";
   sc.clear();
   sc.resize(prods.size());

   for (int p=0; p<prods.size(); ++p) {
      //cout<<p<<"\n";
      //cout<<prods[p]<<"\n";
    //  cout<<pos[p].size()<<" "<<neg[p].size()<<"\n";
      sc[p].train(pos[p], neg[p], prods[p]);
      //cout<<p<<"\n";
   }
   scTrained=true;
   //cout<<"Classifiers trained.\n";
}

void semanticParser::trainClassifiers_semi(const vector <vector <pn> > &un) {
   cout<<"Semi-supervised training classifiers...\n";
   sc.clear();
   sc.resize(prods.size());

   for (int p=0; p<prods.size(); ++p) {
      //cout<<p<<"\n";
      //cout<<prods[p]<<"\n";
    //  cout<<pos[p].size()<<" "<<neg[p].size()<<"\n";
      sc[p].train_semi(pos[p], neg[p], un[p], prods[p]);
      //cout<<p<<"\n";
   }
   scTrained=true;
   cout<<"Semi-supervised classifiers trained.\n";
}


void semanticParser::train(const vector <example> &eg, const vector <production> &ps) {

   if (eg.size() > 0) MAX_MEM /= eg.size();

   DEBUG=false;
   
   trainEg = eg;
   prods = ps;

   getInitialPosNeg();
   //for (int p=0; p<prods.size(); ++p) {
     // if (initPos[p].size() > 0) assert(initPos[p][0].i>=0);
   //}
	 for (int p=0; p<initNeg.size(); ++p) { // CH
	    if (initNeg[p].size() > 1000) {
	       cout<<initNeg[p].size()<<" -> ";
	       randomize_me(initNeg[p]);
	       initNeg[p].resize(1000);
	       cout<<initNeg[p].size()<<"\n";
	    }
	 }

   trainInitClassifiers();

   markRec(recTr, trainEg);

   sc = initSc;
   pos = initPos;
   neg = initNeg;
   
   cout<<"Training...\n";
   int prevIncorr=trainEg.size();
   int incorr=trainEg.size();
   //for (int itr=0; itr<ITR; ++itr) {
   int c=0;
   int maxitr;
   if (ANN+2 < trainEg.size()/20) maxitr = ANN+2;
   else {
      maxitr = trainEg.size()/20;
   }
   int limincorr = trainEg.size()/20; // 20
   while (c < maxitr && incorr > limincorr) {   
      ++c;
      incorr = testOnTrain();
      //cout<<incorr<<" "<<prevIncorr<<"\n";
      //if (incorr >= prevIncorr) break;
      if (incorr <= limincorr) break;
      collectPosNeg();
      trainClassifiers();
      prevIncorr=incorr;
   }
   cout<<"done\n";

   trained = true;
   MAX_MEM *= eg.size();
}


void semanticParser::extract_unlabeled(const example &eg, const vector <vector <int> > &q, vector <vector <pn> > &un) const {
   for (int p=0; p<prods.size(); ++p) {
      if (prods[p].learnable) {
	 for (int a=0; a<q[p].size(); a += 2) {
	    vector <int> uu;
	    for (int b=q[p][a]; b<=q[p][a+1]; ++b) {
	       uu.push_back(eg.nli[b]);
	    }
	    int f=0;
	    for (f=0; f<nli.size(); ++f) {
	       if (nli[f]==uu) break;
	    }
	    if (f < nli.size()) {
	       int e;
	       for (e=0; e<un[p].size(); ++e) {
		  if (f==un[p][e].i) break;
	       }
	       if (e==un[p].size()) un[p].push_back(pn(f));
	    }
	    else {
	       nli.push_back(uu);
	       un[p].push_back(pn(nli.size()-1));
	    }
	 }
      }
   }
}

void semanticParser::train_semi(const vector <example> &eg, vector <example> &unEg, const vector <production> &ps) {

   MAX_MEM /= eg.size();

   DEBUG=false;
   
   trainEg = eg;
   prods = ps;

   getInitialPosNeg();

   trainInitClassifiers();

   markRec(recTr, trainEg);

   sc = initSc;
   pos = initPos;
   neg = initNeg;
   cout<<"Training...\n";
   int prevIncorr=trainEg.size();
   int incorr=trainEg.size();
   //for (int itr=0; itr<ITR; ++itr) {
   int c=0;
   int maxitr;
   if (ANN+2 < trainEg.size()/20) maxitr = ANN+2;
   else maxitr = trainEg.size()/20;
   int limincorr = trainEg.size()/20;
   while (c < maxitr && incorr > limincorr) {   
      ++c;
      incorr = testOnTrain();
      cout<<incorr<<" "<<prevIncorr<<"\n";
      //if (incorr >= prevIncorr) break;
      if (incorr <= limincorr) break;
      collectPosNeg();
      trainClassifiers();
      prevIncorr=incorr;
   }
   cout<<"done\n";

   trained = true;
   MAX_MEM *= eg.size();

   // utilize the unlabeled examples

   vector <vector <pn> > un;
   un.resize(prods.size());

   vector <vector <bool> > rt;
   markRec(rt,unEg);

   RECORD=true;
   for (int i=0; i<unEg.size(); ++i) {
      unEg[i].ans.clear();
#ifdef _CLANG
     if ( 
	   unEg[i].skip || unEg[i].id==38||unEg[i].id==152||unEg[i].id==157||unEg[i].id==392||unEg[i].id==411||unEg[i].id==461
	   ) {
	continue;
     }
#endif
        q.clear();
        q.resize(prods.size());
	earleyParses(START, unEg[i], rt[i],1,0);
        extract_unlabeled(unEg[i], q, un);
   }
   RECORD=false;

   trainClassifiers_semi(un);
}



void semanticParser::test(vector <example> &testEg) const {
   assert(trained);
   MAX_MEM /= testEg.size();

   DEBUG=true;

   vector <vector <bool> > rt;
   markRec(rt,testEg);
   for (int i=0; i<testEg.size(); ++i) {
      testEg[i].ans.clear();
#ifdef _CLANG
     if ( 
	   testEg[i].skip || testEg[i].id==38||testEg[i].id==152||testEg[i].id==157||testEg[i].id==392||testEg[i].id==411||testEg[i].id==461
	   ) {
	continue;
     }
#endif
	earleyParses(START, testEg[i], rt[i],1,0);
   }
   MAX_MEM *= testEg.size();
}

void semanticParser::test_force(vector <example> &testEg) const {
   assert(trained);
   MAX_MEM /= testEg.size();

   DEBUG=true;

   vector <vector <bool> > rt;
   markRec(rt,testEg);
   for (int i=0; i<testEg.size(); ++i) {
      testEg[i].ans.clear();
#ifdef _CLANG
     if ( 
	   testEg[i].skip || testEg[i].id==38||testEg[i].id==152||testEg[i].id==157||testEg[i].id==392||testEg[i].id==411||testEg[i].id==461
	   ) {
	continue;
     }
#endif
	earleyParses(START, testEg[i], rt[i],1,0);
   }
   MAX_MEM *= testEg.size();
}
