#ifndef _SEMANTICPARSER_H
#define _SEMANTICPARSER_H

#include "example.h"
#include "earleyState.h"
#include "stringClassifier.h"

class nega {
   // an intermediate structure to store negative examples 
   public:
      int p, s, e; //production, start, end
      nega (int pp, int ss, int ee) {
	 p = pp; s = ss; e = ee;
      }
      bool operator==(const nega &o) {
	 return p==o.p && s==o.s && e==o.e;
      }
};

class semanticParser {

   // data
   bool trained, scTrained, initScTrained;
   vector <example> trainEg;
   vector <vector <bool> > recTr;
   vector <vector <pn> > pos, neg, initPos, initNeg;
   vector <production> prods;
   vector <stringClassifier> sc, initSc;

   void markRec(vector <vector <bool> > &rec, vector <example> &eg) const;

   // pos-neg
   void collectPosNeg();
   void collectNegFromPos();
   void collectPosEg(const clang &c, int i);
   void collectPosNegEg(int i);
   void collectNegEg(clang &c, const clang &corr, int i);
   bool collectNegHelper(clang &c, const clang &corr, vector <bool> &prob) const;
   bool bfs(vector <clang*> c, vector <clang*> corr, int q, vector <bool> &prob) const;
   void wordCovered(int w, const clang &c, vector <int> &cov, vector <int> &ids)  const;
   void wordCovered(int w, const clang &c, vector <int> &cov) const; 
   void getNegs(const clang &c, vector <nega> &ngs )const;

   // earley's 
   void earleyParses_helper(const string &root, example &teg, const vector <bool> &rec, int beam, double corrLim, bool force) const;
   void scanner(int w, vector <vector <earleyState> > &chart, int end, int ind, example &teg, vector <vector <vector <double> > > &probs, const vector <bool> &rec, bool force) const;
   void predictor(int w, vector <vector <earleyState> > &chart, int end, int ind, example &teg, vector <vector <vector <double> > > &probs, const vector <bool> &rec, bool force) const;
   void completer(int w, vector <vector <earleyState> > &chart, int end, int ind, example &teg, vector <vector <vector <double> > > &probs, const vector <bool> &rec, int beam, double corrLim, bool force)const;
   clang recover(const earleyState &es, const vector <vector <earleyState> > &cs, int depth=0) const;

   bool insert(const earleyState &n, int begin, int end, int nt, vector <vector <vector <earleyBeam> > > &)const;
   void changeES(earleyState &e, vector <vector <earleyState> > &es, vector <vector <vector <earleyBeam> > > &eb, double ol, double ne, int ii)const;
   void findMinBeam(int w, int pc, int begin, vector <vector <earleyState> > &es,  vector <vector <vector <earleyBeam> > > &eb, int ii) const;
   bool exists(const earleyState &n,  const vector <vector <earleyState> > &es, const vector <vector <vector <vector <int> > > > &alleb) const;
   double findProb(int p, int s, int e, example &teg, double &prob) const;
   bool constantString (const string &s) const;
   
   // train-test
   int testOnTrain();
   void getInitialPosNeg();
   void trainInitClassifiers();
   void trainClassifiers();
   void extract_unlabeled(const example &eg, const vector <vector <int> > &q, vector <vector <pn> > &un) const;
   void trainClassifiers_semi(const vector <vector <pn> > &un);
   
   public:

      semanticParser() {
	 trained = false;
	 initScTrained = false;
	 scTrained = false;
      }
      double findProb(int p, int s, int e, example &teg) const;
   
      void earleyParses(const string &root, example &teg, const vector <bool> &rec, int beam, double corrLim, bool force=false) const;

      void train(const vector <example> &eg, const vector <production> &ps);
      void train_semi(const vector <example> &eg, vector <example> &unEg, const vector <production> &ps);

      void test(vector <example> &testEg) const;
      void test_force(vector <example> &testEg) const;
};

#endif
