#include <hash_map.h>
#include <map>
#include "svm_kernel_classifier.h"
#include "common_seq.h"

extern vector <vector <int> > nli;
extern map <string, int> s2n; // mapping from strings to numbers
extern vector <string> n2s; // reverse mapping
extern int STR_KERNEL_LEN;
extern int MAX_MEM2;

extern int ANN;

//vector <vector <double> > k;
vector <hash_map <int, double> > h(MAX_MEM2);
vector <hash_map <int, double> > hn(MAX_MEM2);
//hash_map <int, double> h;
//hash_map <int, double> hn;

#define Malloc(type,n) (type *)malloc((n)*sizeof(type))
/*
inline double svm_kernel_classifier::findKernelUn(int i, int j) const { 
   if (k[i][j] == -1) {
      k[i][j] = ncsUn(nli[i], nli[j],STR_KERNEL_LEN);
      k[j][i] = k[i][j];
   }
   return k[i][j];
}
*/

double svm_kernel_classifier::findKernelUn(int i, int j) const {
   if (i > j) {
      swap(i,j);
   }
   if (i > h.size()) h.resize(2*h.size());
   hash_map <int, double>::iterator itr=h[i].find(j);
   if (itr==h[i].end()) {
      if (h[i].size() > MAX_MEM2) {
	 h[i].clear();
      }
      double d = ncsUn(nli[i], nli[j],STR_KERNEL_LEN);
      h[i][j] =  d;
      return d;
   }
   else {
      return itr->second;
   }
}
double svm_kernel_classifier::findKernel(int i, int j) const {
   if (i > j) {
      swap(i,j);
   }
   if (i > hn.size()) hn.resize(2*hn.size());
   hash_map <int, double>::iterator itr=hn[i].find(j);
   if (itr==hn[i].end()) {
      if (hn[i].size() > MAX_MEM2) {
	 hn[i].clear();
      }
      double deno = sqrt(findKernelUn(i,i)*findKernelUn(j,j));
      assert(deno != 0);
      double d = findKernelUn(i,j)/deno;
      hn[i][j] =  d;
      return d;
   }
   else {
      return itr->second;
   }
}

/*
double svm_kernel_classifier::findKernelUn(int i, int j) const {
   if (h.size() > MAX_MEM2) h.clear();
   int ind;
   if (j < i) {
      ind = i*100000 + j;
   }
   else {
      ind = j*100000 + i;
   }
   if (ind < 0) {
      cout<<i<<" "<<j<<" yes\n";
      assert(ind>=0);
   }
   hash_map <int, double>::iterator itr=h.find(ind);
   if (itr==h.end()) {
      double d = ncsUn(nli[i], nli[j],STR_KERNEL_LEN);
      h[ind] =  d;
      //h[j*10000000+i] = d; // symmetry
      return d;
   }
   else {
      return itr->second;
   }
}
double svm_kernel_classifier::findKernel(int i, int j) const {
   if (i<0 || j<0) {
      cout<<i<<" "<<j<<"\n";
   }
   assert(i >= 0 && j >= 0);
   //assert(i<100000 && j < 100000);
   if (hn.size() > MAX_MEM2) hn.clear();
   int ind;
   if (j < i) {
      ind = j*100000 + i;
   }
   else {
      ind = i*100000 + j;
   }
   if (ind < 0) {
      cout<<i<<" "<<j<<" yes\n";
      assert(ind>=0);
   }
   hash_map <int, double>::iterator itr=hn.find(ind);
   if (itr==hn.end()) {
      double d = findKernelUn(i,j)/sqrt(findKernelUn(i,i)*findKernelUn(j,j));
      //cout<<i<<" "<<j<<" "<<ind<<" - "<<d<<" "<<hn.size()<<"\n";
      hn[ind] =  d;
      //cout<<ind<<" --\n";
      //hn[j*10000000+i] = d;
      return d;
   }
   else {
      return itr->second;
   }
}*/


void svm_kernel_classifier::train(const vector <pn> &ppos, const vector <pn> &pneg) {
   pos = ppos;
   neg = pneg;

   svm_problem prob;
   
   prob.l = pos.size()+neg.size(); 
   prob.y.resize(prob.l);
   //prob.y = Malloc(double,prob.l);
   prob.W = Malloc(double,prob.l);
   prob.x = Malloc(struct svm_node*, prob.l);

   //cout<<prob.l<<" prob.l\n";
   for (int i=0; i<pos.size(); ++i) {
      prob.y[i] = 1;
      prob.W[i] = pos[i].w;
      prob.x[i] = Malloc(struct svm_node, prob.l+2);
      prob.x[i][0].index = 0;
      prob.x[i][0].value = i+1;
      for (int j=0; j<pos.size(); ++j) {
	 prob.x[i][j+1].index = j;
	 prob.x[i][j+1].value = findKernel(pos[i].i, pos[j].i);
      }
      for (int j=0; j<neg.size(); ++j) {
	 prob.x[i][j+pos.size()+1].index = j+pos.size();
	 prob.x[i][j+pos.size()+1].value = findKernel(pos[i].i, neg[j].i);
      }
      prob.x[i][prob.l+1].index = -1;
   }

   for (int i=0; i<neg.size(); ++i) {
      int a=i+pos.size();
      prob.y[a] = -1;
      prob.W[a] = neg[i].w;
      prob.x[a] = Malloc(struct svm_node, prob.l+2);
      prob.x[a][0].index = 0;
      prob.x[a][0].value = a+1;
      for (int j=0; j<pos.size(); ++j) {
	 prob.x[a][j+1].index = j;
	 prob.x[a][j+1].value = findKernel(neg[i].i, pos[j].i);
      }
      for (int j=0; j<neg.size(); ++j) {
	 prob.x[a][j+pos.size()+1].index = j+pos.size();
	 prob.x[a][j+pos.size()+1].value = findKernel(neg[i].i, neg[j].i);
      }
      prob.x[a][prob.l+1].index = -1;
   }

   param.svm_type = C_SVC;
   param.kernel_type = PRECOMPUTED;
   param.degree = 3;
   param.gamma = 0;	// 1/k
   param.coef0 = 0;
   param.nu = 0.5;
   param.cache_size = 100;
   param.C = ANN;//100;
   param.eps = 1e-3;
   param.p = 0.1;
   param.shrinking = 1;
   param.probability = 1; // not default
   param.nr_weight = 0;
   param.weight_label = NULL;
   param.weight = NULL;

   /*
   model = svm_train(&prob, &param);
   svm_save_model("o", model);
   prob.y.clear();
   //free(prob.y);
   for (int j=0; j<prob.l; ++j) {
      free(prob.x[j]);
   }
   free(prob.x);
   svm_destroy_model(model);
   model = svm_load_model("o");*/

   struct svm_model *tmodel = svm_train(&prob,&param);
   model = svm_copy_model(tmodel);
   prob.y.clear();
   free(prob.W);
   for (int j=0; j<prob.l; ++j) {
      free(prob.x[j]);
   }
   free(prob.x);

   l = prob.l;

   svm_destroy_model(tmodel);
   trained = true;

   /*
   int t=0;
   int c=0;
   for (int i=0; i<pos.size(); ++i) {
      if (giveDec(pos[i].i) > 0) ++c;
      ++t;
   }
   for (int i=0; i<neg.size(); ++i) {
      if (giveDec(neg[i].i) <= 0) ++c;
      ++t;
   }
   if (c==0) {
   for (int i=0; i<neg.size(); ++i) {
      cout<<giveDec(neg[i].i)<<" ";
   }
   cout<<"\n";
   }
   cout<<c*100/t<<"% "<<c<<"/"<<t<<" p="<<pos.size()<<" n="<<neg.size()<<"\n";
   */
}

void svm_kernel_classifier::train_semi(const vector <pn> &pppos, const vector <pn> &ppneg, const vector <pn> &puneg) {
   // uses ptsvm for semi-supervised training

   vector <pn> ppos=pppos, pneg=ppneg, uneg=puneg;

   // del will be true only for examples coming from unlabeled
   for (int i=0; i<ppos.size(); ++i) {
      ppos[i].del = false;
   }
   for (int i=0; i<pneg.size(); ++i) {
      pneg[i].del = false;
   }

   train(ppos,pneg);


   int c=0;
   while (c<25 && uneg.size() > 0) {
   //while (c<100 && uneg.size() > 0) { // para ??
      ++c;
      //cout<<"looping\n";
      // dynamical adjusting
      int i=0;
      while (i<ppos.size()) {
	 if (ppos[i].del) {
	    if (giveDec(ppos[i].i) < 0) {
	       uneg.push_back(ppos[i]);
	 /*cout<<"Unlabeling pos: ";
	 for (int a=0; a<nli[ppos[i].i].size(); ++a) {
	    cout<<n2s[nli[ppos[i].i][a]]<<" ";
	 }
	 cout<<"\n";
	 */
	       ppos.erase(ppos.begin()+i);
	    }
	    else ++i;
	 }
	 else ++i;
      }
      i=0;
      while (i<pneg.size()) {
	 if (pneg[i].del) {  
	    if (giveDec(pneg[i].i) > 0) {
	       uneg.push_back(pneg[i]);
	       /*
	 cout<<"Unlabeling neg: ";
	 for (int a=0; a<nli[pneg[i].i].size(); ++a) {
	    cout<<n2s[nli[pneg[i].i][a]]<<" ";
	 }
	 cout<<"\n";
	 */
	       pneg.erase(pneg.begin()+i);
	    }
	    else ++i;
	 }
	 else ++i;
      }
      int a=0;
      while (a < uneg.size()) {
         double d = giveDec(uneg[a].i);
         if (d >=1 || d <=-1) {
	    uneg.erase(uneg.begin()+a);
         }
         else ++a;
      }
      
      double pv=0, nv=0;
      int pi, ni;
      for (int i=0; i<uneg.size(); ++i) {
         double dec = giveDec(uneg[i].i);
         if (dec > 0) {
	    if (dec > pv && dec < 1) {
	       pv = dec;
	       pi = i;
	    }
         }
         if (dec < 0) {
	    if (dec < nv && dec > -1) {
	       nv = dec;
	       ni = i;
	    }
         }
      }
      if (pv > 0) {
	 uneg[pi].del = true;
	 ppos.push_back(uneg[pi]);
	 /*
	 cout<<"Adding pos: ";
	 for (int a=0; a<nli[uneg[pi].i].size(); ++a) {
	    cout<<n2s[nli[uneg[pi].i][a]]<<" ";
	 }
	 cout<<"\n";
	 */
	 uneg.erase(uneg.begin()+pi);
	 if (nv < 0 && pi < ni) --ni;
      }
      if (nv < 0) {
	 uneg[ni].del = true;
	 pneg.push_back(uneg[ni]);
	 /*
	 cout<<"Adding neg: ";
	 for (int a=0; a<nli[uneg[ni].i].size(); ++a) {
	    cout<<n2s[nli[uneg[ni].i][a]]<<" ";
	 }
	 cout<<"\n";
	 */
	 uneg.erase(uneg.begin()+ni);
      }
      train(ppos,pneg);
   }
   //cout<<c<<"\n";
}


/*

double svm_kernel_classifier::giveP(const vector <int> &st) const {
   return giveP(0,st.size()-1,st);
}

double svm_kernel_classifier::giveP(int s, int e, const vector <int> &st) const {
   svm_node *x;
   x = Malloc(struct svm_node, prob.l+1);
   for (int j=0; j<se.sn.size(); ++j) {
      x[j] = se.sn[j];
   }
   x[se.sn.size()].index = -1;

   int nr_class=svm_get_nr_class(model);
   double *prob_estimates;
   prob_estimates = (double *) malloc(nr_class*sizeof(double));

   double c = svm_predict_probability(model,x,prob_estimates);

   double r;

	 if (c==1) {
	    if (nr_class==2) {
	       r = max(prob_estimates[0], prob_estimates[1]);
	    }
	    else {
	       r =  prob_estimates[0];
	    }
	 }
	 else {
	    if (nr_class==2) {
	       r =  min(prob_estimates[0], prob_estimates[1]);
	    }
	    else {
	       r =  1-prob_estimates[0];
	    }
	 }
	 free(x);
	 free(prob_estimates);
	 if (r==0) {
	    r = 0.000001; 
	 }
   return r;
}
*/
double svm_kernel_classifier::giveP(int s, int e, example &eg) const {
   assert(trained);
   assert(s<=e);
   
   svm_node *x;
   x = Malloc(struct svm_node, l+2);
   x[0].index = 0;
   for (int i=0; i<pos.size(); ++i) {
      x[i+1].index = i;
      double deno = sqrt(eg.selfK[s][e]*findKernelUn(pos[i].i,pos[i].i));
      assert(deno != 0);
      x[i+1].value = eg.giveK(s,e,pos[i].i,nli[pos[i].i])/deno;
   }
   for (int i=0; i<neg.size(); ++i) {
      int a = i+pos.size();
      x[a+1].index = a;
      double deno = sqrt(eg.selfK[s][e]*findKernelUn(neg[i].i,neg[i].i));
      assert(deno != 0);
      x[a+1].value = eg.giveK(s,e,neg[i].i,nli[neg[i].i])/deno;
   }
      //if (eg.k.size() > 10000 && eg.k.size() % 10000 == 0) cout<<eg.k.size()<<" <------\n";
   x[l+1].index = -1;

   int nr_class=svm_get_nr_class(model);
   double *prob_estimates;
   prob_estimates = (double *) malloc(nr_class*sizeof(double));

   double c = svm_predict_probability(model,x,prob_estimates);
   double r;
   if (c==1) {
      if (nr_class==2) {
	 r = max(prob_estimates[0], prob_estimates[1]);
      }
      else {
	 r =  prob_estimates[0];
      }
   }
   else {
      if (nr_class==2) {
	 r =  min(prob_estimates[0], prob_estimates[1]);
      }
      else {
	 r =  1-prob_estimates[0];
      }
   }
   free(x);
   free(prob_estimates);
   if (r==0) {
      r = 0.000001; 
   }

   //cout<<r<<" "<<giveDec(s,e,eg)<<"\n";
   return r;
}


double svm_kernel_classifier::giveDec(int s, int e, example &eg) const {
   assert(trained);
   assert(s<=e);
   
   svm_node *x;
   x = Malloc(struct svm_node, l+2);
   x[0].index = 0;
   for (int i=0; i<pos.size(); ++i) {
      x[i+1].index = i;
      double deno = sqrt(eg.selfK[s][e]*findKernelUn(pos[i].i,pos[i].i));
      assert(deno != 0);
      x[i+1].value = eg.giveK(s,e,pos[i].i,nli[pos[i].i])/deno;
   }
   for (int i=0; i<neg.size(); ++i) {
      int a = i+pos.size();
      x[a+1].index = a;
      double deno = sqrt(eg.selfK[s][e]*findKernelUn(neg[i].i,neg[i].i));
      assert(deno != 0);
      x[a+1].value = eg.giveK(s,e,neg[i].i,nli[neg[i].i])/deno;
   }
      //if (eg.k.size() > 10000 && eg.k.size() % 10000 == 0) cout<<eg.k.size()<<" <------\n";
   x[l+1].index = -1;

   int nr_class=svm_get_nr_class(model);
   double *dec_values;
   dec_values = (double *) malloc((nr_class*(nr_class-1)/2)*sizeof(double));
   svm_predict_values(model,x,dec_values);

   double r;
   
   if (model->label[0]==1) {
      r = dec_values[0];
   }
   else {
      r = 0-dec_values[0];
   }
   
   free(x);
   free(dec_values);

   return r;
}


double svm_kernel_classifier::giveDec(int u) const {
   // gives decision value for the example nli[u]
   assert(trained);
   
   svm_node *x;
   x = Malloc(struct svm_node, l+2);
   x[0].index = 0;
   for (int i=0; i<pos.size(); ++i) {
      x[i+1].index = i;
      x[i+1].value = findKernel(u,pos[i].i);
   }
   for (int i=0; i<neg.size(); ++i) {
      int a = i+pos.size();
      x[a+1].index = a;
      x[a+1].value = findKernel(u, neg[i].i);
   }
      
   x[l+1].index = -1;

   int nr_class=svm_get_nr_class(model);
   double *dec_values;
   dec_values = (double *) malloc((nr_class*(nr_class-1)/2)*sizeof(double));
   svm_predict_values(model,x,dec_values);

   double r;
   
   if (model->label[0]==1) {
      r = dec_values[0];
   }
   else {
      r = 0-dec_values[0];
   }
   
   free(x);
   free(dec_values);

   return r;
}
