#ifndef _SVM_KERNEL_CLASSIFIER_H
#define _SVM_KERNEL_CLASSIFIER_H

#include "pn.h"
#include "example.h"
#include "svm.h"


extern bool DEBUG;

#define Malloc(type,n) (type *)malloc((n)*sizeof(type))


class svm_kernel_classifier {
   bool trained;

   int l;
   svm_parameter param;
   svm_model *model;

   vector <pn> pos, neg;

   double findKernelUn(int i, int j) const;
   double findKernel(int i, int j) const;

   public:
       svm_kernel_classifier() {
	  trained = false;
	  model= NULL;
       }
       
       svm_kernel_classifier (const svm_kernel_classifier &o) {
	  trained = o.trained;
	  model = NULL;
	  if (! trained) return;
	  l = o.l;
	  pos = o.pos;
	  neg = o.neg;
	  param = o.param;
	  model = svm_copy_model(o.model);
       }

       void operator=(const svm_kernel_classifier &o) {
	  trained = o.trained;
	  if (! trained) return;
	  l = o.l;
	  pos = o.pos;
	  neg = o.neg;
	  param = o.param;
	  model = svm_copy_model(o.model);
       }
       
      ~svm_kernel_classifier () {
	 if (trained) {
	    //cout<<"Destroying..\n";
	    svm_destroy_model(model);
	    svm_destroy_param(&param);
	    //cout<<"Destroyed\n";
	 }
      }
      void train(const vector <pn> &pos, const vector <pn> &neg);
      void train_semi(const vector <pn> &pos, const vector <pn> &neg, const vector <pn> &uneg);
      //double giveP(const vector <int> &st) const;
      //double giveP(int s, int e, const vector <int> &st) const;
      double giveP(int s, int e, example &eg) const;
      double giveDec(int s, int e, example &eg) const;
      double giveDec(int i) const;
};

#endif
