#include <cmath>
#include "tree.h"
#include "util.h"
#include "production.h"

extern vector <vector <string> > synProds; // syntactic productions
extern vector <production> prods; // syntactic productions

bool ctree::preTerm() {
   if (preTermComputed) return preTerminal;
   else {
      if (child.size()==0) {
	 preTerminal = false;
	 preTermComputed = true;
	 return false;
      }
      preTerminal = true;
      for (int c=0; c<child.size(); ++c) {
         if (child[c]->child.size() > 0) {
	    preTerminal = false;
	    preTermComputed = true;
	    return false;
         }
      }
      preTermComputed = true;
      return true;
   }
}

void ctree::getProds() {
   vector <string> p;
   p.push_back(root);
   if (child.size() != 0) {
      for (int i=0; i<child.size(); ++i) {
         p.push_back(child[i]->root);
      }
   }
   rootProd = -1;
   for (int i=0; i<synProds.size(); ++i) {
      if (p==synProds[i]) {
	 rootProd = i;
	 break;
      }
   }
   if (rootProd == -1) {
      synProds.push_back(p);
      rootProd = synProds.size()-1;
   }
   for (int i=0; i<child.size(); ++i) {
      child[i]->getProds();
   }
}

int ctree::numberLeaves(int i) {
   if (child.size()==0) {
      n = i;
      return i+1;
   }
   else {
      sleaf = i;
      for (int c=0; c<child.size(); ++c) {
	 i = child[c]->numberLeaves(i);
      }
      eleaf = i-1;
      return i;
   }
}

int ctree::findSize() {
   size = 1;
   for (int i=0; i<child.size(); ++i) size += child[i]->findSize();
   return size;
}

int ctree::assignNodeIds(int i) {
   nodeId = i;
   ++i;
   for (int c=0; c<child.size(); ++c) {
      i = child[c]->assignNodeIds(i);
   }
   return i;
}

void ctree::assignDepth(int i) {
   d = i;
   for (int c=0; c<child.size(); ++c) {
      child[c]->assignDepth(i+1);
   }
}

void ctree::init() {
   numberLeaves(0);
   //collapseOneChild();
   getProds();
   findSize();
   assignNodeIds(0);
   assignDepth(0);
   sk = -1;
}

bool ctree::read(ifstream &in) {
   in>>noskipws;
   bool r = readHelp(in);
   in>>skipws;
   init();
   return r;
}
bool ctree::read(FILE *in) {
   bool r = readHelp(in);
   init();
   return r;
}

bool ctree::readHelp(ifstream &in) {
   root = "";
   char c;
   in>>c;
   while (c==' '||c=='\n'||c=='\t') {
      in>>c;
   }
   if (c=='(') {
      in>>c;
      while (c != ' ' && c!= '\n' && c != '\t' && c!=')' && c!='(') {
	 root.push_back(c);
	 in>>c;
      }
      while (true) { // reading children
         while (c==' '||c=='\n'||c=='\t') {
            in>>c;
         }
         in.putback(c);
         if (c==')') {
	    in>>c;
	    return true;
	 }

         ctree *ch=new(ctree);
         if (ch->readHelp(in)) child.push_back(ch);
         in>>c;
      }
   }
   else {
      // can only be a leaf
      while (c != ' ' && c!= '\n' && c != '\t' && c!=')' && c!='(') {
	 root.push_back(c);
	 in>>c;
      }
      if (c=='(' || c==')') {
         in.putback(c);
      }
      return true;
   }
   if (root=="") return false;
}
bool ctree::readHelp(FILE *in) {
   root = "";
   char c;
   c=getc(in);
   while (c==' '||c=='\n'||c=='\t') {
      c=getc(in);
   }
   if (c=='(') {
      c=getc(in);
      while (c != ' ' && c!= '\n' && c != '\t' && c!=')' && c!='(') {
	 root.push_back(c);
         c=getc(in);
      }
      while (true) { // reading children
         while (c==' '||c=='\n'||c=='\t') {
            c=getc(in);
         }
         ungetc(c, in);
         if (c==')') {
            c=getc(in);
	    return true;
	 }

         ctree *ch=new(ctree);
         if (ch->readHelp(in)) child.push_back(ch);
	 c=getc(in);
      }
   }
   else {
      // can only be a leaf
      while (c != ' ' && c!= '\n' && c != '\t' && c!=')' && c!='(') {
	 root.push_back(c);
	 c=getc(in);
      }
      if (c=='(' || c==')') {
         ungetc(c, in);
      }
      return true;
   }
   if (root=="") return false;
}

ctree ctree::buildCover(int s, int e) const {
   // assumes it already covers leaves from s to e
   ctree t;
   t.root = root;
   //t.child.resize(child.size());
   for (int i=0; i<child.size(); ++i) {
      if (child[i]->sleaf > e || child[i]->eleaf < s) {
	 t.child.push_back(new ctree(child[i]->root));
      }
      else {
	  t.child.push_back(new ctree(child[i]->buildCover(s,e)));
      }
   }
   return t;
}
   
ctree ctree::leafCover(int s, int e) const {
   // returns subtree covering the leaves from s to e

   assert(child.size()>0 && covers(s,e));

   for (int i=0; i<child.size(); ++i) {
      if (child[i]->child.size() > 0 && child[i]->covers(s,e)) {
	 return child[i]->leafCover(s,e);
      }
   }

   ctree t = buildCover(s,e);
   t.init();
   return t;
}

string ctree::firstAncestor(int s, int e) const {
   // returns the first ancestor for leaves from s to e
   assert(child.size()>0 && covers(s,e));
   for (int i=0; i<child.size(); ++i) {
      if (child[i]->child.size() > 0 && child[i]->covers(s,e)) {
	 return child[i]->firstAncestor(s,e);
      }
   }
   return root;
}

void ctree::leafString(vector <string> &ls) const {
   // its string of leaves 
   if (child.size() > 0) {
      for (int i=0; i<child.size(); ++i) {
	 child[i]->leafString(ls);
      }
   }
   else {
      ls.push_back(root);
   }
}

void ctree::posEgs(vector <int> &p, vector <int> &sl, vector <int> &el) {
   numberLeaves(0);
   posEgs_helper(p,sl,el);
}
void ctree::posEgs_helper(vector <int> &p, vector <int> &sl, vector <int> &el) const {
   // positive examples from the mrapt (start and end leaves)
   if (rootMRProd != -1) {
      p.push_back(rootMRProd);
      sl.push_back(sleaf);
      el.push_back(eleaf);
   }
   for (int i=0; i<child.size(); ++i) {
      if (child[i]->child.size() > 0) {
	 if (child[i]->rootMRProd==rootMRProd) {
	    //child[i]->rootMRProd = -1;
	    child[i]->sleaf = sleaf;
	    child[i]->eleaf = eleaf;
	 }
	 child[i]->posEgs_helper(p,sl,el);
      }
   }
}

ostream& ctree::write(ostream &out) const {
   if (child.size() > 0) {
      //out<<"("<<root<<" "; 
      out<<"("<<root<<" "<<sleaf<<" "<<eleaf<<" "; 
   }
   //else out<<root<<" ";
   else out<<root<<" "<<sleaf<<" "<<eleaf<<" ";
   for (int i=0; i<child.size(); ++i) {
      if (i != 0) out<<" ";
      out<<*child[i];
   }
   if (child.size() > 0) out<<")";
   return out;
}


ostream& ctree::writeMR(ostream &out) const {
   if (child.size() > 0) {
      out<<"(";
      if (rootMRProd != -1) {
	 out<<prods[rootMRProd]<<" ";
      }
   }
   for (int i=0; i<child.size(); ++i) {
      if (child[i] != NULL) child[i]->writeMR(out);
   }
   if (child.size() > 0) out<<")\n";
   return out;
}

void makeNt(ctree &t, vector <NT> &pnt) {
   if (t.child.size() == 0) return;
   NT n;
   n.nodeId = t.nodeId;
   n.rootProd = t.rootProd;
   n.ptr = &t;
   pnt.push_back(n);

   for (int c=0; c<t.child.size(); ++c) {
      makeNt(*t.child[c],pnt);
   }
}

void makeNt(ctree &t) {
   t.nt.clear();
   makeNt(t, t.nt);
   t.ntComp();
}


double delta(ctree& n1, ctree &n2, vector <vector <double> > &c, vector <vector <bool> > &done)  {

   assert(n1.child.size() > 0 && n2.child.size() > 0);

   if (done[n1.nodeId][n2.nodeId]) return c[n1.nodeId][n2.nodeId];

   if (n1.preTerm() && n2.preTerm()) {
      c[n1.nodeId][n2.nodeId] = TREE_LAMBDA;
      done[n1.nodeId][n2.nodeId] = true;
      return TREE_LAMBDA;
   }

   /*/ orginal kernel
   if (n1.rootProd != n2.rootProd) {
      c[n1.nodeId][n2.nodeId] = 0;
      done[n1.nodeId][n2.nodeId] = true;
      return 0;
   }
   double r = TREE_LAMBDA;
   assert(n1.child.size()==n2.child.size());
   for (int i=0; i<n1.child.size(); ++i) {
      if (n1.child[i]->child.size() > 0 && n2.child[i]->child.size() > 0)
         r *= 1 + delta(*n1.child[i], *n2.child[i], c, done);
   }*/


   // modified kernel
   if (synProds[n1.rootProd][0] != synProds[n2.rootProd][0]) {
      c[n1.nodeId][n2.nodeId] = 0;
      done[n1.nodeId][n2.nodeId] = true;
      return 0;
   }
   double match=0;
   int m=0;
   double r=1;
   for (int i=0; i<n1.child.size(); ++i) {
      for (int j=m; j<n2.child.size(); ++j) {
	 if (n1.child[i]->root==n2.child[j]->root) {
	    m = j+1;
	    ++match;
            if (n1.child[i]->child.size() > 0 && n2.child[j]->child.size() > 0)
	       r *= 1 + delta(*n1.child[i], *n2.child[j], c, done);
	 }
      }
   }
   r *= TREE_LAMBDA;
   //r *= TREE_LAMBDA/((match*match)/((synProds[n1.rootProd].size()-1)*(synProds[n2.rootProd].size()-1)));
   

   c[n1.nodeId][n2.nodeId] = r;
   done[n1.nodeId][n2.nodeId] = true;
   return r; 
}



double kernelUn(ctree &t1, ctree &t2)  {
   vector <vector <double> > c(t1.size, vector <double> (t2.size, 0));
   vector <vector <bool> > done(t1.size, vector <bool> (t2.size, false));

   // find evaluate_set_pair
   int i1=0, i2=0;
   vector <NT> m1, m2; // matched nodes

   if (! t1.ifNtComp()) makeNt(t1);
   if (! t2.ifNtComp()) makeNt(t2);

   while (i1 < t1.nt.size() && i2 < t2.nt.size()) {
      if (t1.nt[i1].rootProd > t2.nt[i2].rootProd) ++i2;
      else {
	 if (t1.nt[i1].rootProd < t2.nt[i2].rootProd) ++i1;
	 else {
	    int si2 = i2;
	    while (t1.nt[i1].rootProd == t2.nt[i2].rootProd && i1 < t1.nt.size()) {
	       while (t1.nt[i1].rootProd == t2.nt[i2].rootProd && i2 < t2.nt.size()) {
		  m1.push_back(t1.nt[i1]);
		  m2.push_back(t2.nt[i2]);
		  ++i2;
	       }
	       ++i1;
	       i2 = si2;
	    }
	 }
      }
   }
   
   double r=0;
   for (int i=0; i<m1.size(); ++i) {
      if (m1[i].ptr->child.size() > 0 && m2[i].ptr->child.size() > 0)
         r += delta(*m1[i].ptr, *m2[i].ptr, c, done);
   }
   return r;
   //return 1+r*r;
}

double selfKernel(ctree &t1) {
   return kernelUn(t1, t1);
}

double kernel(ctree &t1, ctree &t2) {
   double r = kernelUn(t1, t2);
   if (r==0) return 0;
   else return r/sqrt(selfKernel(t1)*selfKernel(t2));
}
