#include "treeKernel.h"

double computeC(int i, int j,  vector <vector <double> > &c, vector <vector <double> > &done, const  vector <nodeInfo> &ni1,  const vector <nodeInfo> &ni2) {
   
   if (done[i][j]) return c[i][j];

   if (!ni1[i].complete || !ni2[j].complete || ni1[i].prod != ni2[j].prod) {
      // nodes have different productions
      c[i][j] = 0;
      done[i][j] = true;
   }
   else {
      if (ni1[i].childIndex.size()==0 || ni2[j].childIndex.size()==0) {
	 // nodes are leaves
	 c[i][j] = TREE_LAMBDA;
	 done[i][j] = true;
      }
      else {
	 // recursive case
	 double r=TREE_LAMBDA;
	 if (ni2[j].childIndex.size()!=ni1[i].childIndex.size()) {
	    cout<<ni1[i]<<" "; 
	    cout<<ni1[i].childIndex.size()<<" ->\n";
	    cout<<ni2[j]<<" "; 
	    cout<<ni2[j].childIndex.size()<<" ->\n";
	 }
	 assert(ni2[j].childIndex.size()==ni1[i].childIndex.size());
	 for (int e=0; e<ni1[i].childIndex.size(); ++e) {
	    if (ni1[i].childIndex[e] != -1 && ni2[j].childIndex[e] != -1)
	       r *= 1 + computeC(ni1[i].childIndex[e], ni2[j].childIndex[e], c, done, ni1, ni2);
	 }
	 c[i][j] = r;
	 done[i][j] = true;
      }
   }
   return c[i][j];
}

double treeKernelUn(const vector <nodeInfo> &ni1, const vector <nodeInfo> &ni2) {
   // unnormalized tree kernel

   vector <vector <double> > c(ni1.size(), vector <double> (ni2.size(), 0));
   vector <vector <double> > done(ni1.size(), vector <double> (ni2.size(), false));

   double r=0;
   for (int i=0; i<ni1.size(); ++i) {
      for (int j=0; j<ni2.size(); ++j) {
	 r += computeC(i, j, c, done, ni1, ni2);
      }
   }
   return r;
}
double treeKernel(const vector <nodeInfo> &ni1, const vector <nodeInfo> &ni2) {
   return treeKernelUn(ni1,ni2)/sqrt(treeKernelUn(ni1,ni1)*treeKernelUn(ni2,ni2));
}


