#include <iostream>
#include <cmath>
#include <fstream>
#include "util.h"

#include <algorithm>
#include <vector>
#include <map>

using namespace std;

extern map <string, int> s2n; // mapping from strings to numbers

ngram::ngram (const string &ss, const vector <int> &wwnn) {
	 s = ss;
	 wn = wwnn;
	 map<string, int>::iterator itr=s2n.find(s);
         if (itr==s2n.end()) {
	    s2n[s] = n2s.size();
	    n = n2s.size();
	    n2s.push_back(s);
	 }
         else {
	    n = itr->second;
	 }
}

string lowerCase(const string &s) {
   string r=s;
   for (int i=0; i<s.size(); ++i) {
      if (s[i] >= 65 && s[i] <= 90)
      r[i] = s[i] + 32;
   }
   return r;
}

// converts string to double
double stod(const string &s) {
   return atof(s.c_str());
}

double power(double a, int n) {
   double r=1;
   for (int i=0; i<n; ++i) r *= a;
   return r;
}


// returns the string after converting all its upper case letters to lower case
std::string decap(std::string s) {
   for (int i=0; i<s.size(); ++i) {
      if (s[i]>=65 && s[i]<=90) s[i] += 32;
   }
   return s;
}

/*
// removes stop words
void removeStopWords(vector <string> &s) {
   static vector <string> sw;

   if (sw.size()==0) {
      ifstream in("/u/rjkate/krisp/stopwords.txt");
      //ifstream in("/u/mooney/ir-code/ir/utilities/stopwords.txt");
      while (in) {
	 string tmp;
	 in>>tmp;
	 if (in) {
	    sw.push_back(tmp);
	 }
      }
   }

   
   int i=0;
   while (i<s.size()) {
      if (find(sw.begin(), sw.end(), s[i]) != sw.end()) {
	 s.erase(s.begin()+i);
      }
      else ++i;
   }
}*/

// pushes all the n-grams of s on snew 
void nGramize(const vector <string> &s, vector <string> &snew, int n) {
   assert(n>0);
   if (s.size() < n) return;
   for (int i=0; i<=s.size()-n; ++i) {
      string ng=s[i];
      for (int j=1; j<n; ++j) {
	 ng += " " + s[i+j];
      }
      snew.push_back(ng);
   }
}
// pushes all the n-grams of s on snew 
void nGramize(const vector <string> &s, vector <ngram> &snew, int n) {
   assert(n>0);
   if (s.size() < n) return;
   for (int i=0; i<=s.size()-n; ++i) {
      vector <int> w;
      string ng=s[i];
      w.push_back(i);
      for (int j=1; j<n; ++j) {
	 ng += " " + s[i+j];
         w.push_back(i+j);
      }
      snew.push_back(ngram(ng,w));
   }
}


void nGramize_separate(const vector <string> &s, vector <vector <string> > &snew, vector <vector <int> > &nnew, int n) {
   assert(n>0);
   if (s.size() < n) return;
   for (int i=0; i<=s.size()-n; ++i) {
      vector <string> ng;
      vector <int> nn;
      ng.push_back(s[i]);
      nn.push_back(i);
      for (int j=1; j<n; ++j) {
	 ng.push_back(s[i+j]);
	 nn.push_back(i+j);
      }
      snew.push_back(ng);
      nnew.push_back(nn);
   }
}

long fact(int n) {
   long r=1;
   for (int i=2; i<=n; ++i) r*= i;
   return r;
}

int nC(int n, int a) {
   assert(n>=a);
   if (n > 2*a) a = n-a;
   long r=1;
   for (int i=a+1; i<=n; ++i) r *= i;
   return r/fact(n-a);
   //return fact(n)/(fact(a)*fact(n-a));
}



void permu(const vector <int> &a, vector <bool> &u, vector <vector <int> > &b, int k) {
   if (k > 0) {
      for (int i=0; i<a.size(); ++i) {
         if (u[i]) {
	    u[i]=false;
	    if (k==1) {
	       b.push_back(vector <int> (1, a[i]));
	    }
	    else {
	       int s=b.size();
	       permu(a, u, b, k-1);
	       for (int j=s; j<b.size(); ++j) {
	          b[j].push_back(a[i]);
	       }
	    }
	    u[i] = true;
	 }
      }
   }
}

void permu(const vector <int> &a, vector <vector <int> > &b, int k) {
   vector <bool> u(a.size(), true);
   permu(a, u, b, k);
}

void combi(const vector <int> &a, vector <bool> &u, vector <vector <int> > &b, int k, int p=0) {
   if (k > 0) {
      for (int i=p; i<a.size(); ++i) {
         if (u[i]) {
	    u[i]=false;
	    if (k==1) {
	       b.push_back(vector <int> (1, a[i]));
	    }
	    else {
	       int s=b.size();
	       combi(a, u, b, k-1, i+1);
	       for (int j=s; j<b.size(); ++j) {
	          b[j].push_back(a[i]);
	       }
	    }
	    u[i] = true;
	 }
      }
   }
}


void combi(const vector <int> &a, vector <vector <int> > &b, int k) {
   vector <bool> u(a.size(), true);
   combi(a, u, b, k);
}

// pushes all the n-grams of s with total gap g on snew 
void nGramize_with_gaps(const vector <string> &s, vector <ngram> &snew, int n, int g) {
   assert(n>0 && g>=0);

   if (s.size() < n+g) return;

   vector <vector <string> > ss;
   vector <vector <int> > nn;
   nGramize_separate(s, ss, nn, n+g);

   
   vector <int> c(n+g-2);
   for (int i=0; i<c.size(); ++i) {
      c[i] = i+1;
   }
   vector <vector <int> > cc;
   combi(c, cc, n-2);

   for (int i=0; i<ss.size(); ++i) {

      for (int j=0; j<cc.size(); ++j) {
	 vector <int> w;
	 int ind=0;
	 string s=ss[i][0];
	 w.push_back(nn[i][0]);

	 for (int k=0; k<cc[j].size(); ++k) {
	    for (int e=ind; e<cc[j][k]-1; ++e) s.append("_"); 
	    s += ' ';
	    s.append(ss[i][cc[j][k]]);
	    w.push_back(nn[i][cc[j][k]]);
	    ind = cc[j][k];
	 }
	 for (int e=ind; e<ss[i].size()-2; ++e) s.append("_"); 
	 s += ' ';
	 s.append(ss[i][ss[i].size()-1]);
	 w.push_back(nn[i][ss[i].size()-1]);
	 snew.push_back(ngram(s,w));
      }
   }
}

int getS2N(const string &s) {
   map <string, int>::iterator itr=s2n.find(s);
   if (itr==s2n.end()) {
      int r;
      s2n[s] = n2s.size();
      r = n2s.size();
      n2s.push_back(s);
      return r;
   }   
   else return itr->second;
}


double dotProduct(const vector <double> &a, const vector <double> &b) {
   assert(a.size()==b.size());

   double r=0;

   for (int i=0; i<a.size(); ++i) r += a[i]*b[i];

   return sqrt(r);
}
double cosine(const vector <double> &a, const vector <double> &b) {
   double da=dotProduct(a,a), db=dotProduct(b,b);
   if (da==0 || db==0) return 0;
   else return dotProduct(a,b)/(da*db);
}

void swap (vector <int> &a, int i, int j) {
   if (i == j) return;
   int tmp = a[i];
   a[i] = a[j];
   a[j] = tmp;
}
	
void reverse (vector <int> &a, int b, int e) {
   while (b < e) {
      --e;
      swap(a, b, e);
      ++b;
   }
}

    
bool nextPermutation (vector <int> &a) {
   // makes the next permutation if possible
   int b=0, e=a.size();
   for (int i=e - 1, j=e; i != b;) {
      --i;
      --j;
      if (a[i] < a[j]) {
	 int k = e;
	 while (!(a[i] < a[--k])) {}
	 swap(a, i, k);
	 reverse(a, j, e);
	 return true;
      }
   }
   reverse(a, b, e);
   return false;
}

double weightedSelfKernel (const vector <double> &a, int n) {
   // finds all the subsequences of length at most n 
   // returns the score of all the weighted squared subsequences

   if (n < 1) return 0;
   
   double r=0;
   for (int i=0; i<a.size(); ++i) {
      r += a[i]*a[i];
   }

   for (int i=2; a.size() <= i && i<=n; ++i) {
      vector <int> p(i); // to store permutations
      for (int j=0; j<p.size(); ++j) p[j]=j;

      bool per=true;
      while (per) {
	 
	 double s=0;
	 for (int j=0; j<p.size(); ++j) {
	    s += a[p[j]];
	 }

	 r += s*s;
	 
	 per = nextPermutation(p);
      }
   }
   return r;
}

double maxWord(const vector <double> &a, const vector <double> &b) {
   assert(a.size()==b.size());

   double max=-999999;

   for (int i=0; i<a.size(); ++i) {
      if (a[i]*b[i] > max) {
	 max = a[i]*b[i];
      }
   }
 
   return sqrt(max);
}


double similar(const vector <double> &a, const vector <double> &b, int &mi) {

   assert(a.size()==b.size());
   
   double max=0;
   mi=0;
   for (int i=0; i<a.size(); ++i) {
      if (a[i]*b[i] > max) {
	 max = a[i]*b[i];
	 mi = i;
      }
   }
   return max;
   
   /*
   int ai=-1, bi=-1;
   double max=-999999;
   for (int i=0; i<a.size(); ++i) {
      if (a[i] > max) {
	 max = a[i];
	 ai = i;
      }
   }
   max=-999999;
   for (int i=0; i<b.size(); ++i) {
      if (b[i] > max) {
	 max = b[i];
	 bi = i;
      }
   }

   return 1-((bi-ai)*(bi-ai)*1.0)/(a.size()*b.size());*/

   /*
   double aa=0, bb=0;
   double sa=0, sb=0;
   
   for (int i=0; i<a.size(); ++i) {
      aa += i*a[i];
      sa += a[i];
   }
   for (int i=0; i<b.size(); ++i) {
      bb += i*b[i];
      sb += b[i];
   }

   double ava; 
   if (sa != 0) ava = aa/sa;
   else ava = 0;
   double avb;
   if (sb != 0) avb = bb/sb;
   else avb = 0;


   //cout<<1-(ava-avb)/a.size()<<ava<<": "<<bb<<" "<<sb<<" "<<avb<<" "<<(ava-avb)<<" "<<" "<<a.size()<<" ----\n";
      
   if (a.size()==0) return 0;

   if (ava > avb) return 1-(ava-avb)/a.size();
   else  return 1-(avb-ava)/a.size();*/
}
