#ifndef _UTIL_H
#define _UTIL_H

#include <vector>
#include <string>
#include <sstream>
#include <cassert>
#include <sys/types.h>
#include <unistd.h>

using namespace std;

#define _PC
//#define _GEO

extern vector <string> n2s; // reverse mapping

class ngram {
   public: 
      string s;
      int n;
      vector <int> wn;

      bool operator < (const ngram &other) const {
	 return n < other.n;
      }
      bool operator > (const ngram &other) const {
	 return n > other.n;
      }
      bool operator == (const ngram &other) const {
	 return n == other.n;
      }

      ngram (const string &ss, const vector <int> &wwnn);
      ngram (int i) {
	 n = i;
	 s = n2s[i];
      }
};

template <typename T>
T sq(const T &a) {
   return a*a;
}

double power(double a, int n);

long fact(int n);
int nC(int n, int a);

// used to convert an int or a double to string
template <typename T>
string itos(T n) {  
   std::ostringstream ost;
   ost<<n;
   return ost.str();
}

// converts string to double
double stod(const string &s);

// writes the vector
template <typename T>
void writeVector(ostream &out, const vector <T> &v) {
   for (int i=0; i<v.size(); ++i) {
      out<<v[i]<<" ";
   }
   out<<"\n";
}

// returns the minimum element in the vector v
template <typename T>
T my_min_element(const vector <T> &v) {
   assert(v.size() >= 0);
   T r = v[0];
   for (int i=1; i<v.size(); ++i) {
      if (v[i] < r) r = v[i];
   }
   return r;
}

// randomizes or randomly permutes the vector, returns the randomized vector
template <typename T>
vector<T> randomize(vector <T> arg) {
  // srandom(getpid());
   static int seed = 0;
   srandom(seed++);
   T temp;
   for (int i=0; i<arg.size(); ++i) {
      int r = 0;
      if (i > 0) r = random()%i;
      temp = arg[i];
      arg[i] = arg[r];
      arg[r] = temp;
   }
   return arg;
}

// randomizes or randomly permutes the vector itself 
template <typename T>
void randomize_me(vector <T> &arg) {
  // srandom(getpid());
   static int seed = getpid();
   //cout<<seed<<"\n";
   //exit(1);
   srandom(seed++);
   T temp;
   for (int i=0; i<arg.size(); ++i) {
      int r = 0;
      if (i > 0) r = random()%i;
      temp = arg[i];
      arg[i] = arg[r];
      arg[r] = temp;
   }
}

// removes duplicates from the vector
template <typename T>
void remove_duplicates(vector <T> &a) {
   int i=1;
   while (i < a.size()) {
      bool rem=false;
      for (int j=0; j<i; ++j) {
	 if (a[i]==a[j]) {
	    rem = true;
	    break;
	 }
      }
      if (rem) {
	 a.erase(a.begin()+i);
      }
      else ++i;
   }
}

// return intersection of the elements of a and b 
template <typename T>
vector <T> intersection(vector <T> a, const vector <T> &b) {
   vector <T> c;
   remove_duplicates(a);
   
   for (int i=0; i<a.size(); ++i) {
      if (find(b.begin(), b.end(), a[i]) != b.end()) {
	 c.push_back(a[i]);
      }
   }
   return c;
}

int getS2N(const string &s);
double dotProduct(const vector <double> &a, const vector <double> &b);
double cosine(const vector <double> &a, const vector <double> &b);
double maxWord(const vector <double> &a, const vector <double> &b);
double similar(const vector <double> &a, const vector <double> &b, int &mi);

double weightedSelfKernel (const vector <double> &a, int n);

// returns the string after converting all its upper case letters to lower case
std::string decap(std::string s);

// removes stop words
void removeStopWords(vector <string> &);

// selects k items from a
void combi(const vector <int> &a, vector <vector <int> > &b, int k);
// selects with permutation k items from a
void permu(const vector <int> &a, vector <vector <int> > &b, int k);

// pushes all the n-grams of s on snew 
void nGramize(const vector <string> &s, vector <string> &snew, int n);
// pushes all the n-grams of s on snew 
void nGramize(const vector <string> &s, vector <ngram> &snew, int n);
// pushes all the n-grams of s with total gap g on snew 
void nGramize_with_gaps(const vector <string> &s, vector <ngram> &snew, int n, int g);


#endif
